# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from decimal import Decimal
from unittest.mock import Mock

import pytest

from braket.ahs.analog_hamiltonian_simulation import DiscretizationError
from braket.ahs.atom_arrangement import AtomArrangement, SiteType


@pytest.fixture
def default_atom_arrangement():
    atom_arrangement = (
        AtomArrangement()
        .add((0, 3.1e-6), SiteType.FILLED)
        .add((0, 5.99e-6))
        .add((3.001e-6, 0))
        .add((3e-6, 3e-6))
        .add((-3.01e-6, 6.5e-6))
    )
    return atom_arrangement


def test_add_chaining():
    atom_arrangement = (
        AtomArrangement()
        .add(coordinate=(0, 0), site_type=SiteType.FILLED)
        .add((0, 3), SiteType.FILLED)
        .add((0, 6))
        .add((3, 0))
        .add((3, 3))
        .add((3, 6))
        .add((3, 0))
    )
    assert len(atom_arrangement) == 7


def test_iteration():
    values = [(0, 0), (0.1, 0.2), (Decimal(0.3), Decimal(0.4))]
    atom_arrangement = AtomArrangement()
    for value in values:
        atom_arrangement.add(value)
    returned_values = []
    for site in atom_arrangement:
        returned_values.append(site.coordinate)
    assert values == returned_values


def test_coordinate_list():
    values = [(0, 0), (0.1, 0.2), (Decimal(0.3), Decimal(0.4))]
    atom_arrangement = AtomArrangement()
    for value in values:
        atom_arrangement.add(value)
    for coord_index in range(2):
        coords = atom_arrangement.coordinate_list(coord_index)
        assert coords == [value[coord_index] for value in values]


@pytest.mark.parametrize(
    "position_res, expected_x, expected_y",
    [
        # default x: [0, 0, 3.001e-6, 3e-6, -3.01e-6]
        # default y: [3.1e-6, 5.99e-6, 0, 3e-6, 6.5e-6]
        (
            Decimal("1e-6"),
            [Decimal("0"), Decimal("0"), Decimal("3e-6"), Decimal("3e-6"), Decimal("-3e-6")],
            [Decimal("3e-6"), Decimal("6e-6"), Decimal("0"), Decimal("3e-6"), Decimal("6e-6")],
        ),
        (
            Decimal("1e-7"),
            [Decimal("0"), Decimal("0"), Decimal("3e-6"), Decimal("3e-6"), Decimal("-3e-6")],
            [Decimal("3.1e-6"), Decimal("6e-6"), Decimal("0"), Decimal("3e-6"), Decimal("6.5e-6")],
        ),
        (
            Decimal("2e-7"),
            [Decimal("0"), Decimal("0"), Decimal("3e-6"), Decimal("3e-6"), Decimal("-3e-6")],
            [Decimal("3e-6"), Decimal("6e-6"), Decimal("0"), Decimal("3e-6"), Decimal("6.4e-6")],
        ),
        (
            Decimal("1e-8"),
            [Decimal("0"), Decimal("0"), Decimal("3e-6"), Decimal("3e-6"), Decimal("-3.01e-6")],
            [
                Decimal("3.1e-6"),
                Decimal("5.99e-6"),
                Decimal("0"),
                Decimal("3e-6"),
                Decimal("6.5e-6"),
            ],
        ),
    ],
)
def test_discretize(default_atom_arrangement, position_res, expected_x, expected_y):
    properties = Mock()
    properties.lattice.geometry.positionResolution = position_res
    actual = default_atom_arrangement.discretize(properties)
    assert expected_x == actual.coordinate_list(0)
    assert expected_y == actual.coordinate_list(1)


@pytest.mark.xfail(raises=DiscretizationError)
def test_invalid_discretization_properties(default_atom_arrangement):
    properties = "not-a-valid-discretization-property"
    default_atom_arrangement.discretize(properties)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Iterator
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from numbers import Number
from typing import Union

import numpy as np

from braket.ahs.discretization_types import DiscretizationError, DiscretizationProperties


class SiteType(Enum):
    VACANT = "Vacant"
    FILLED = "Filled"


@dataclass
class AtomArrangementItem:
    """Represents an item (coordinate and metadata) in an atom arrangement."""

    coordinate: tuple[Number, Number]
    site_type: SiteType

    def _validate_coordinate(self) -> None:
        if len(self.coordinate) != 2:
            raise ValueError(f"{self.coordinate} must be of length 2")
        for idx, num in enumerate(self.coordinate):
            if not isinstance(num, Number):
                raise TypeError(f"{num} at position {idx} must be a number")

    def _validate_site_type(self) -> None:
        allowed_site_types = {SiteType.FILLED, SiteType.VACANT}
        if self.site_type not in allowed_site_types:
            raise ValueError(f"{self.site_type} must be one of {allowed_site_types}")

    def __post_init__(self) -> None:
        self._validate_coordinate()
        self._validate_site_type()


class AtomArrangement:
    def __init__(self):
        """Represents a set of coordinates that can be used as a register to an
        AnalogHamiltonianSimulation.
        """
        self._sites = []

    def add(
        self,
        coordinate: Union[tuple[Number, Number], np.ndarray],
        site_type: SiteType = SiteType.FILLED,
    ) -> AtomArrangement:
        """Add a coordinate to the atom arrangement.

        Args:
            coordinate (Union[tuple[Number, Number], ndarray]): The coordinate of the
                atom (in meters). The coordinates can be a numpy array of shape (2,)
                or a tuple of int, float, Decimal
            site_type (SiteType): The type of site. Optional. Default is FILLED.

        Returns:
            AtomArrangement: returns self (to allow for chaining).
        """
        self._sites.append(AtomArrangementItem(tuple(coordinate), site_type))
        return self

    def coordinate_list(self, coordinate_index: Number) -> list[Number]:
        """Returns all the coordinates at the given index.

        Args:
            coordinate_index (Number): The index to get for each coordinate.

        Returns:
            list[Number]:The list of coordinates at the given index.

        Example:
            To get a list of all x-coordinates: coordinate_list(0)
            To get a list of all y-coordinates: coordinate_list(1)
        """
        return [site.coordinate[coordinate_index] for site in self._sites]

    def __iter__(self) -> Iterator:
        return self._sites.__iter__()

    def __len__(self):
        return self._sites.__len__()

    def discretize(self, properties: DiscretizationProperties) -> AtomArrangement:
        """Creates a discretized version of the atom arrangement,
        rounding all site coordinates to the closest multiple of the
        resolution. The types of the sites are unchanged.

        Args:
            properties (DiscretizationProperties): Capabilities of a device that represent the
                resolution with which the device can implement the parameters.

        Raises:
            DiscretizationError: If unable to discretize the program.

        Returns:
            AtomArrangement: A new discretized atom arrangement.
        """
        try:
            position_res = properties.lattice.geometry.positionResolution
            discretized_arrangement = AtomArrangement()
            for site in self._sites:
                new_coordinates = tuple(
                    round(Decimal(c) / position_res) * position_res for c in site.coordinate
                )
                discretized_arrangement.add(new_coordinates, site.site_type)
            return discretized_arrangement
        except Exception as e:
            raise DiscretizationError(f"Failed to discretize register {e}")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import time
from logging import Logger, getLogger
from typing import Any, Optional, Union

from braket.aws.aws_session import AwsSession
from braket.jobs.metrics_data.definitions import MetricStatistic, MetricType
from braket.jobs.metrics_data.exceptions import MetricsRetrievalError
from braket.jobs.metrics_data.log_metrics_parser import LogMetricsParser


class CwlInsightsMetricsFetcher:
    LOG_GROUP_NAME = "/aws/braket/jobs"
    QUERY_DEFAULT_JOB_DURATION = 3 * 60 * 60

    def __init__(
        self,
        aws_session: AwsSession,
        poll_timeout_seconds: float = 10,
        poll_interval_seconds: float = 1,
        logger: Logger = getLogger(__name__),
    ):
        """Initializes a `CwlInsightsMetricsFetcher`.

        Args:
            aws_session (AwsSession): AwsSession to connect to AWS with.
            poll_timeout_seconds (float): The polling timeout for retrieving the metrics,
                in seconds. Default: 10 seconds.
            poll_interval_seconds (float): The interval of time, in seconds, between polling
                for results. Default: 1 second.
            logger (Logger): Logger object with which to write logs, such as quantum task statuses
                while waiting for a quantum task to be in a terminal state. Default is
                `getLogger(__name__)`
        """
        self._poll_timeout_seconds = poll_timeout_seconds
        self._poll_interval_seconds = poll_interval_seconds
        self._logger = logger
        self._logs_client = aws_session.logs_client

    @staticmethod
    def _get_element_from_log_line(
        element_name: str, log_line: list[dict[str, Any]]
    ) -> Optional[str]:
        """Finds and returns an element of a log line from CloudWatch Insights results.

        Args:
            element_name (str): The element to find.
            log_line (list[dict[str, Any]]): An iterator for RegEx matches on a log line.

        Returns:
            Optional[str]: The value of the element with the element name, or None if no such
            element is found.
        """
        return next(
            (element["value"] for element in log_line if element["field"] == element_name), None
        )

    def _get_metrics_results_sync(self, query_id: str) -> list[Any]:
        """Waits for the CloudWatch Insights query to complete and then returns all the results.

        Args:
            query_id (str): CloudWatch Insights query ID.

        Raises:
            MetricsRetrievalError: Raised if the query is Failed or Cancelled.

        Returns:
            list[Any]: The results from CloudWatch insights 'GetQueryResults' operation.
        """
        timeout_time = time.time() + self._poll_timeout_seconds
        while time.time() < timeout_time:
            response = self._logs_client.get_query_results(queryId=query_id)
            query_status = response["status"]
            if query_status in ["Failed", "Cancelled"]:
                raise MetricsRetrievalError(f"Query {query_id} failed with status {query_status}.")
            elif query_status == "Complete":
                return response["results"]
            else:
                time.sleep(self._poll_interval_seconds)
        self._logger.warning(f"Timed out waiting for query {query_id}.")
        return []

    def _parse_log_line(self, result_entry: list[dict[str, Any]], parser: LogMetricsParser) -> None:
        """Parses the single entry from CloudWatch Insights results and adds any metrics it finds
        to 'all_metrics' along with the timestamp for the entry.

        Args:
            result_entry (list[dict[str, Any]]): A structured result from calling CloudWatch
                Insights to get logs that contain metrics. A single entry contains the message
                (the actual line logged to output), the timestamp (generated by CloudWatch Logs),
                and other metadata that we (currently) do not use.
            parser (LogMetricsParser) : The CWL metrics parser.
        """
        message = self._get_element_from_log_line("@message", result_entry)
        if message:
            timestamp = self._get_element_from_log_line("@timestamp", result_entry)
            parser.parse_log_message(timestamp, message)

    def _parse_log_query_results(
        self, results: list[Any], metric_type: MetricType, statistic: MetricStatistic
    ) -> dict[str, list[Union[str, float, int]]]:
        """Parses CloudWatch Insights results and returns all found metrics.

        Args:
            results (list[Any]): A structured result from calling CloudWatch Insights to get
                logs that contain metrics.
            metric_type (MetricType): The type of metrics to get.
            statistic (MetricStatistic): The statistic to determine which metric value to use
                when there is a conflict.

        Returns:
            dict[str, list[Union[str, float, int]]]: The metrics data.
        """
        parser = LogMetricsParser()
        for result in results:
            self._parse_log_line(result, parser)
        return parser.get_parsed_metrics(metric_type, statistic)

    def get_metrics_for_job(
        self,
        job_name: str,
        metric_type: MetricType = MetricType.TIMESTAMP,
        statistic: MetricStatistic = MetricStatistic.MAX,
        job_start_time: int | None = None,
        job_end_time: int | None = None,
        stream_prefix: str | None = None,
    ) -> dict[str, list[Union[str, float, int]]]:
        """Synchronously retrieves all the algorithm metrics logged by a given Hybrid Job.

        Args:
            job_name (str): The name of the Hybrid Job. The name must be exact to ensure only the
                relevant metrics are retrieved.
            metric_type (MetricType): The type of metrics to get. Default is MetricType.TIMESTAMP.
            statistic (MetricStatistic): The statistic to determine which metric value to use
                when there is a conflict. Default is MetricStatistic.MAX.
            job_start_time (int | None): The time when the hybrid job started.
                Default: 3 hours before job_end_time.
            job_end_time (int | None): If the hybrid job is complete, this should be the time at
                which the hybrid job finished. Default: current time.
            stream_prefix (str | None): If a logs prefix is provided, it will be used instead
                of the job name.

        Returns:
            dict[str, list[Union[str, float, int]]]: The metrics data, where the keys
            are the column names and the values are a list containing the values in each row.

        Example:
            timestamp energy
            0         0.1
            1         0.2
            would be represented as:
            { "timestamp" : [0, 1], "energy" : [0.1, 0.2] }
            The values may be integers, floats, strings or None.
        """
        query_end_time = job_end_time or int(time.time())
        query_start_time = job_start_time or query_end_time - self.QUERY_DEFAULT_JOB_DURATION

        stream_prefix = (stream_prefix or job_name).replace("/", "\\/")

        query = (
            f"fields @timestamp, @message "
            f"| filter @logStream like /^{stream_prefix}\\// "
            f"| filter @message like /Metrics - /"
        )

        response = self._logs_client.start_query(
            logGroupName=self.LOG_GROUP_NAME,
            startTime=query_start_time,
            endTime=query_end_time,
            queryString=query,
            limit=10000,
        )

        query_id = response["queryId"]

        results = self._get_metrics_results_sync(query_id)

        return self._parse_log_query_results(results, metric_type, statistic)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import ABC, abstractmethod
from typing import Any


class Operator(ABC):
    """An operator is the abstract definition of an operation for a quantum device."""

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the operator.

        Returns:
            str: The name of the operator.
        """

    @abstractmethod
    def to_ir(self, *args, **kwargs) -> Any:
        """Converts the operator into the canonical intermediate representation.
        If the operator is passed in a request, this method is called before it is passed.

        Returns:
            Any: The the canonical intermediate representation of the operator.
        """



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from braket import ipython_utils
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import (  # noqa: F401
    AnalogHamiltonianSimulationQuantumTaskResult,
    AnalogHamiltonianSimulationShotStatus,
)
from braket.tasks.annealing_quantum_task_result import AnnealingQuantumTaskResult  # noqa: F401
from braket.tasks.gate_model_quantum_task_result import GateModelQuantumTaskResult  # noqa: F401
from braket.tasks.photonic_model_quantum_task_result import (  # noqa: F401
    PhotonicModelQuantumTaskResult,
)
from braket.tasks.quantum_task import QuantumTask  # noqa: F401
from braket.tasks.quantum_task_batch import QuantumTaskBatch  # noqa: F401

# Apply nest_asyncio if currently running within Jupyter. This ensures anything that uses
# asyncio will run in Jupyter without any issues.
#
# Inspired by https://github.com/ipython/ipython/issues/11694 and
# https://github.com/jupyter/notebook/issues/3397#issuecomment-419386811
if ipython_utils.running_in_jupyter():
    import nest_asyncio

    nest_asyncio.apply()



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import abstractmethod

from braket.circuits.noise_model.criteria import Criteria
from braket.registers.qubit_set import QubitSetInput


class InitializationCriteria(Criteria):
    """Criteria that implement these methods may be used to determine initialization noise."""

    @abstractmethod
    def qubit_intersection(self, qubits: QubitSetInput) -> QubitSetInput:
        """Returns subset of passed qubits that match the criteria.

        Args:
            qubits (QubitSetInput): A qubit or set of qubits that may match the criteria.

        Returns:
            QubitSetInput: The subset of passed qubits that match the criteria.
        """
        raise NotImplementedError



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.jobs.metrics_data import LogMetricsParser
from braket.jobs.metrics_data.definitions import MetricStatistic, MetricType

MALFORMED_METRICS_LOG_LINES = [
    {"timestamp": "Test timestamp 0", "message": ""},
    {"timestamp": "Test timestamp 1", "message": "No semicolon metric0=2.0"},
    {"timestamp": "Test timestamp 2", "message": "metric0=not_a_number;"},
    {"timestamp": "Test timestamp 3", "message": "also not a number metric0=2 . 0;"},
    {"timestamp": "Test timestamp 3", "message": "metric0=;"},
    {"timestamp": "Test timestamp 3", "message": "metric0= ;"},
    {"timestamp": "Test timestamp 4"},
    {"unknown": "Unknown"},
]

SIMPLE_METRICS_LOG_LINES = [
    # This is a standard line of what our metrics may look like
    {
        "timestamp": "Test timestamp 0",
        "message": "Metrics - metric0=0.0; metric1=1.0; metric2=2.0 ;",
    },
    # This line overwrites the timestamp by having it output in the metrics.
    {
        "timestamp": "Test timestamp 1",
        "message": "Metrics - timestamp=1628019160; metric0=0.1; metric2= 2.1;",
    },
    # This line adds metric3 that won't have values for any other timestamp
    {
        "timestamp": "Test timestamp 2",
        "message": "Metrics - metric0=0.2; metric1=1.2; metric2= 2.2 ; metric3=0.2;",
    },
    # This line adds metrics expressed as exponents
    {
        "timestamp": "Test timestamp 3",
        "message": "Metrics - metric0=-0.4; metric1=3.14e-22; metric2=3.14E22;",
    },
]

SIMPLE_METRICS_RESULT = {
    "timestamp": [
        "Test timestamp 0",
        1628019160,
        "Test timestamp 2",
        "Test timestamp 3",
    ],
    "metric0": [0.0, 0.1, 0.2, -0.4],
    "metric1": [1.0, None, 1.2, 3.14e-22],
    "metric2": [2.0, 2.1, 2.2, 3.14e22],
    "metric3": [None, None, 0.2, None],
}

MULTINODE_METRICS_LOG_LINES = [
    {
        "timestamp": "Test timestamp 0",
        "message": "[nodeA]<stdout>:Metrics - metric0=1.0;",
    },
    # This line logs the same metric from a different node.
    {
        "timestamp": "Test timestamp 0",
        "message": "[nodeB]<stdout>:Metrics - metric0=2.0;",
    },
    # This line also logs a metric unique to one node.
    {
        "timestamp": "Test timestamp 1",
        "message": "[nodeA]<stdout>:Metrics - metricA=3.0;",
    },
    # This line logs a metric without a node tag.
    {
        "timestamp": "Test timestamp 1",
        "message": "Metrics - metric0=0.0;",
    },
]

MULTINODES_METRICS_RESULT = {
    "timestamp": ["Test timestamp 0", "Test timestamp 0", "Test timestamp 1", "Test timestamp 1"],
    "node_id": [
        "nodeA",
        "nodeB",
        "nodeA",
        None,
    ],
    "metric0": [1.0, 2.0, None, 0.0],
    "metricA": [None, None, 3.0, None],
}

# This will test how metrics are combined when the multiple metrics have the same timestamp
SINGLE_TIMESTAMP_METRICS_LOG_LINES = [
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric0=0.0;"},
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric0=0.1; metric1=1.1;"},
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric0=0.2; metric2=2.8;"},
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric0=0.3; metric1=1.3;"},
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric1=1.4; metric2=2.4;"},
    {
        "timestamp": "Test timestamp 0",
        "message": "Metrics - metric0=0.5; metric1=1.5; metric2=2.5;",
    },
    {"timestamp": "Test timestamp 0", "message": "Metrics - metric1=0.6; metric0=0.6;"},
]


ITERATION_AND_TIMESTAMPS_LOG_LINES = [
    {"timestamp": "Test timestamp 0", "message": "Metrics - iteration_number=0; metric0=0.0;"},
    {
        "timestamp": "Test timestamp 1",
        "message": "Metrics - metric0=0.1; metric1=1.1; iteration_number=0;",
    },
    {"timestamp": "Test timestamp 2", "message": "Metrics - metric0=0.2; metric2=2.8;"},
    {"timestamp": "Test timestamp 3", "message": "Metrics - metric0=0.3; metric1=1.3;"},
    {
        "timestamp": "Test timestamp 4",
        "message": "Metrics - metric1=1.4; metric2=2.4; iteration_number=0;",
    },
    {
        "timestamp": "Test timestamp 5",
        "message": "Metrics - metric0=0.5; metric1=1.5; metric2=2.5;",
    },
    {
        "timestamp": "Test timestamp 6",
        "message": "Metrics - metric1=0.6; iteration_number=0; metric0=0.6;",
    },
]


SINGLE_TIMESTAMP_MAX_RESULTS = {
    "timestamp": ["Test timestamp 0"],
    "metric0": [0.6],
    "metric1": [1.5],
    "metric2": [2.8],
}

SINGLE_TIMESTAMP_MIN_RESULTS = {
    "timestamp": ["Test timestamp 0"],
    "metric0": [0.0],
    "metric1": [0.6],
    "metric2": [2.4],
}

ITERATION_NUMBER_MAX_RESULTS = {
    "iteration_number": [0],
    "timestamp": ["Test timestamp 6"],
    "metric0": [0.6],
    "metric1": [1.4],
    "metric2": [2.4],
}


@pytest.mark.parametrize(
    "log_events, metric_type, metric_stat, metrics_results",
    [
        ([], MetricType.TIMESTAMP, MetricStatistic.MAX, {}),
        (MALFORMED_METRICS_LOG_LINES, MetricType.TIMESTAMP, MetricStatistic.MAX, {}),
        (
            SIMPLE_METRICS_LOG_LINES,
            MetricType.TIMESTAMP,
            MetricStatistic.MAX,
            SIMPLE_METRICS_RESULT,
        ),
        (
            MULTINODE_METRICS_LOG_LINES,
            MetricType.TIMESTAMP,
            MetricStatistic.MAX,
            MULTINODES_METRICS_RESULT,
        ),
        (
            SINGLE_TIMESTAMP_METRICS_LOG_LINES,
            MetricType.TIMESTAMP,
            MetricStatistic.MAX,
            SINGLE_TIMESTAMP_MAX_RESULTS,
        ),
        (
            SINGLE_TIMESTAMP_METRICS_LOG_LINES,
            MetricType.TIMESTAMP,
            MetricStatistic.MIN,
            SINGLE_TIMESTAMP_MIN_RESULTS,
        ),
        (
            ITERATION_AND_TIMESTAMPS_LOG_LINES,
            MetricType.ITERATION_NUMBER,
            MetricStatistic.MAX,
            ITERATION_NUMBER_MAX_RESULTS,
        ),
        # TODO: https://app.asana.com/0/1199668788990775/1200502190825620
        #  We should also test some real-world data, once we have it.
    ],
)
def test_get_all_metrics_complete_results(log_events, metric_type, metric_stat, metrics_results):
    parser = LogMetricsParser()
    for log_event in log_events:
        parser.parse_log_message(log_event.get("timestamp"), log_event.get("message"))
    assert parser.get_parsed_metrics(metric_type, metric_stat) == metrics_results



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from datetime import timedelta
from decimal import Decimal
from unittest.mock import patch

import pytest

from braket.tracking.tracker import Tracker
from braket.tracking.tracking_context import active_trackers
from braket.tracking.tracking_events import (
    _TaskCompletionEvent,
    _TaskCreationEvent,
    _TaskStatusEvent,
)


@pytest.fixture()
def empty_tracker():
    with Tracker() as _tracker:
        yield _tracker


def test_tracker_enter_exit():
    x = len(active_trackers())
    with Tracker() as t:
        assert len(active_trackers()) == 1 + x
        with Tracker() as s:
            assert len(active_trackers()) == 2 + x
            assert s in active_trackers()
            assert t in active_trackers()
    assert len(active_trackers()) == x


def test_tracker_start_stop():
    t = Tracker()
    assert t not in active_trackers()
    t.start()
    assert t in active_trackers()
    t.stop()
    assert t not in active_trackers()


def test_receive_fake_event(empty_tracker):
    event = "NOT AN EVENT"
    with pytest.raises(ValueError):
        empty_tracker.receive_event(event)


CREATE_EVENTS = [
    _TaskCreationEvent(arn="task1:::region", shots=100, is_job_task=True, device="qpu/foo"),
    _TaskCreationEvent(arn="task2:::region", shots=100, is_job_task=False, device="qpu/foo"),
    _TaskCreationEvent(
        arn="job_sim_task:::region", shots=0, is_job_task=True, device="simulator/bar"
    ),
    _TaskCreationEvent(
        arn="notjob_sim_task:::region", shots=0, is_job_task=False, device="simulator/bar"
    ),
    _TaskCreationEvent(
        arn="task_fail:::region", shots=0, is_job_task=False, device="simulator/tn1"
    ),
    _TaskCreationEvent(
        arn="task_cancel:::region", shots=0, is_job_task=False, device="simulator/baz"
    ),
    _TaskCreationEvent(
        arn="2000qtask:::region", shots=100, is_job_task=False, device="qpu/2000Qxyz"
    ),
    _TaskCreationEvent(
        arn="adv_task:::region", shots=100, is_job_task=False, device="qpu/Advantage_system123"
    ),
    _TaskCreationEvent(
        arn="unfinished_sim_task:::region", shots=1000, is_job_task=False, device="simulator/bar"
    ),
    _TaskCreationEvent(
        arn="no_price:::region", shots=1000, is_job_task=False, device="something_else"
    ),
    _TaskCreationEvent(
        arn="unbilled_task0:::region",
        shots=100,
        is_job_task=True,
        device="qpu/foo",
    ),
    _TaskCreationEvent(
        arn="unbilled_task1:::region",
        shots=100,
        is_job_task=True,
        device="qpu/foo",
    ),
]

GET_EVENTS = [
    _TaskStatusEvent(arn="untracked_task:::region", status="FOO"),
    _TaskStatusEvent(arn="task1:::region", status="BAR"),
    _TaskStatusEvent(arn="task2:::region", status="FAILED"),
]
COMPLETE_EVENTS = [
    _TaskCompletionEvent(arn="untracked_task:::region", execution_duration=999999, status="BAR"),
    _TaskCompletionEvent(arn="task1:::region", execution_duration=None, status="COMPLETED"),
    _TaskCompletionEvent(arn="job_sim_task:::region", execution_duration=123, status="COMPLETED"),
    _TaskCompletionEvent(
        arn="notjob_sim_task:::region", execution_duration=1729, status="COMPLETED"
    ),
    _TaskCompletionEvent(arn="task_fail:::region", execution_duration=12345, status="FAILED"),
    _TaskCompletionEvent(arn="task_cancel:::region", execution_duration=None, status="CANCELLED"),
    _TaskCompletionEvent(
        arn="unbilled_task0:::region",
        execution_duration=123,
        status="COMPLETED",
        has_reservation_arn=True,
    ),
    _TaskCompletionEvent(
        arn="unbilled_task1:::region",
        execution_duration=123,
        status="COMPLETED",
        has_reservation_arn=True,
    ),
]


@pytest.fixture
def create_tracker(empty_tracker):
    for e in CREATE_EVENTS:
        empty_tracker.receive_event(e)
    return empty_tracker


@pytest.fixture
def get_tracker(create_tracker):
    for e in GET_EVENTS:
        create_tracker.receive_event(e)
    return create_tracker


@pytest.fixture
def completed_tracker(get_tracker):
    for e in COMPLETE_EVENTS:
        get_tracker.receive_event(e)
    return get_tracker


def test_tracked_resources(create_tracker):
    assert len(CREATE_EVENTS) == len(create_tracker.tracked_resources())


def mock_qpu_price(**kwargs):
    if "Shot" in kwargs["Product Family"]:
        return [{"PricePerUnit": "0.001", "Currency": "USD"}]
    else:
        return [{"PricePerUnit": "1.0", "Currency": "USD"}]


@patch("braket.tracking.tracker.price_search")
def test_qpu_task_cost(price_mock, completed_tracker):
    price_mock.side_effect = mock_qpu_price
    cost = completed_tracker.qpu_tasks_cost()
    assert cost == Decimal("3.3")

    price_mock.side_effect = [[]]
    with pytest.raises(ValueError, match="Found 0 products"):
        completed_tracker.qpu_tasks_cost()

    price_mock.side_effect = [[{}], [{}, {}]]
    with pytest.raises(ValueError, match="Found 2 products"):
        completed_tracker.qpu_tasks_cost()

    price_mock.side_effect = [[{"Currency": "BAD"}], [{"Currency": "BAD"}]]
    with pytest.raises(ValueError, match="Expected USD"):
        completed_tracker.qpu_tasks_cost()


@patch("braket.tracking.tracker.price_search")
def test_simulator_task_cost(price_mock, completed_tracker):
    price_mock.return_value = [{"PricePerUnit": "6.0", "Currency": "USD", "Unit": "minutes"}]
    cost = completed_tracker.simulator_tasks_cost()
    expected = Decimal("0.0001") * (3000 + 3000 + 12345)
    assert cost == expected

    price_mock.return_value = []
    with pytest.raises(ValueError, match="Found 0 products"):
        completed_tracker.simulator_tasks_cost()

    price_mock.return_value = [{"Currency": "BAD"}]
    with pytest.raises(ValueError, match="Expected USD"):
        completed_tracker.simulator_tasks_cost()


def test_quantum_task_statistics(completed_tracker):
    stats = completed_tracker.quantum_tasks_statistics()
    expected = {
        "qpu/foo": {
            "shots": 400,
            "tasks": {"COMPLETED": 3, "FAILED": 1},
            "execution_duration": timedelta(microseconds=246000),
            "billed_execution_duration": timedelta(0),
        },
        "simulator/bar": {
            "shots": 1000,
            "tasks": {"COMPLETED": 2, "CREATED": 1},
            "execution_duration": timedelta(seconds=1, microseconds=852000),
            "billed_execution_duration": timedelta(seconds=6),
        },
        "simulator/tn1": {
            "shots": 0,
            "tasks": {"FAILED": 1},
            "execution_duration": timedelta(seconds=12, microseconds=345000),
            "billed_execution_duration": timedelta(seconds=12, microseconds=345000),
        },
        "simulator/baz": {"shots": 0, "tasks": {"CANCELLED": 1}},
        "qpu/2000Qxyz": {"shots": 100, "tasks": {"CREATED": 1}},
        "qpu/Advantage_system123": {"shots": 100, "tasks": {"CREATED": 1}},
        "something_else": {"shots": 1000, "tasks": {"CREATED": 1}},
    }
    assert stats == expected



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import os
import tempfile
import time
from http.client import HTTPMessage
from pathlib import Path
from unittest.mock import MagicMock, Mock, patch

import boto3
import pytest
from botocore.exceptions import ClientError
from botocore.stub import Stubber

import braket._schemas as braket_schemas
import braket._sdk as braket_sdk
from braket.aws import AwsSession

TEST_S3_OBJ_CONTENTS = {
    "TaskMetadata": {
        "Id": "blah",
    }
}


@pytest.fixture
def boto_session():
    _boto_session = Mock()
    _boto_session.region_name = "us-west-2"
    _boto_session.profile_name = "test-profile"
    return _boto_session


@pytest.fixture
def braket_client():
    _braket_client = Mock()
    _braket_client.meta.region_name = "us-west-2"
    return _braket_client


@pytest.fixture
def aws_session(boto_session, braket_client, account_id):
    _aws_session = AwsSession(boto_session=boto_session, braket_client=braket_client)

    _aws_session._sts = Mock()
    _aws_session._sts.get_caller_identity.return_value = {
        "Account": account_id,
    }
    _aws_session._s3 = Mock()
    return _aws_session


@pytest.fixture
def aws_explicit_session():
    _boto_session = Mock()
    _boto_session.region_name = "us-test-1"
    _boto_session.profile_name = "test-profile"

    creds = Mock()
    creds.access_key = "access key"
    creds.secret_key = "secret key"
    creds.token = "token"
    creds.method = "explicit"
    _boto_session.get_credentials.return_value = creds

    _aws_session = Mock()
    _aws_session.boto_session = _boto_session
    _aws_session._default_bucket = "amazon-braket-us-test-1-00000000"
    _aws_session.default_bucket.return_value = _aws_session._default_bucket
    _aws_session._custom_default_bucket = False
    _aws_session.account_id = "00000000"
    _aws_session.region = "us-test-1"
    return _aws_session


@pytest.fixture
def aws_env_session():
    _boto_session = Mock()
    _boto_session.region_name = "us-test-1"
    _boto_session.profile_name = "test-profile"

    creds = Mock()
    creds.method = "env"
    _boto_session.get_credentials.return_value = creds

    _aws_session = Mock()
    _aws_session.boto_session = _boto_session
    _aws_session._default_bucket = "amazon-braket-us-test-1-00000000"
    _aws_session.default_bucket.return_value = _aws_session._default_bucket
    _aws_session._custom_default_bucket = False
    _aws_session.account_id = "00000000"
    _aws_session.region = "us-test-1"
    return _aws_session


@pytest.fixture
def account_id():
    return "000000000"


@pytest.fixture
def job_role_name():
    return "AmazonBraketJobsExecutionRole-134534514345"


@pytest.fixture
def job_role_arn(job_role_name):
    return f"arn:aws:iam::0000000000:role/{job_role_name}"


@pytest.fixture
def get_job_response():
    return {
        "algorithmSpecification": {
            "scriptModeConfig": {
                "entryPoint": "my_file:start_here",
                "s3Uri": "s3://amazon-braket-jobs/job-path/my_file.py",
            }
        },
        "checkpointConfig": {
            "localPath": "/opt/omega/checkpoints",
            "s3Uri": "s3://amazon-braket-jobs/job-path/checkpoints",
        },
        "instanceConfig": {
            "instanceCount": 1,
            "instanceType": "ml.m5.large",
            "volumeSizeInGb": 1,
        },
        "jobArn": "arn:aws:braket:us-west-2:875981177017:job/job-test-20210628140446",
        "jobName": "job-test-20210628140446",
        "outputDataConfig": {"s3Path": "s3://amazon-braket-jobs/job-path/output"},
        "roleArn": "arn:aws:iam::875981177017:role/AmazonBraketJobRole",
        "status": "RUNNING",
    }


@pytest.fixture
def resource_not_found_response():
    return {
        "Error": {
            "Code": "ResourceNotFoundException",
            "Message": "unit-test-error",
        }
    }


@pytest.fixture
def throttling_response():
    return {
        "Error": {
            "Code": "ThrottlingException",
            "Message": "unit-test-error",
        }
    }


def test_initializes_boto_client_if_required(boto_session):
    AwsSession(boto_session=boto_session)
    boto_session.client.assert_any_call("braket", config=None, endpoint_url=None)


def test_user_supplied_braket_client():
    boto_session = Mock()
    boto_session.region_name = "foobar"
    braket_client = Mock()
    braket_client.meta.region_name = "foobar"
    aws_session = AwsSession(boto_session=boto_session, braket_client=braket_client)
    assert aws_session.braket_client == braket_client


@patch.dict("os.environ", {"BRAKET_ENDPOINT": "some-endpoint"})
def test_config(boto_session):
    config = Mock()
    AwsSession(boto_session=boto_session, config=config)
    boto_session.client.assert_any_call(
        "braket",
        config=config,
        endpoint_url="some-endpoint",
    )


def test_region():
    boto_region = "boto-region"
    braket_region = "braket-region"

    boto_session = Mock()
    boto_session.region_name = boto_region
    braket_client = Mock()
    braket_client.meta.region_name = braket_region

    assert (
        AwsSession(
            boto_session=boto_session,
        ).region
        == boto_region
    )

    assert (
        AwsSession(
            braket_client=braket_client,
        ).region
        == braket_region
    )

    regions_must_match = (
        "Boto Session region and Braket Client region must match and currently "
        "they do not: Boto Session region is 'boto-region', but "
        "Braket Client region is 'braket-region'."
    )
    with pytest.raises(ValueError, match=regions_must_match):
        AwsSession(
            boto_session=boto_session,
            braket_client=braket_client,
        )


def test_iam(aws_session):
    aws_session._iam = Mock()
    assert aws_session.iam_client
    aws_session.boto_session.client.assert_not_called()
    aws_session._iam = None
    assert aws_session.iam_client
    aws_session.boto_session.client.assert_called_with("iam", region_name="us-west-2")


def test_s3(aws_session):
    assert aws_session.s3_client
    aws_session.boto_session.client.assert_not_called()
    aws_session._s3 = None
    assert aws_session.s3_client
    aws_session.boto_session.client.assert_called_with("s3", region_name="us-west-2")


def test_sts(aws_session):
    assert aws_session.sts_client
    aws_session.boto_session.client.assert_not_called()
    aws_session._sts = None
    assert aws_session.sts_client
    aws_session.boto_session.client.assert_called_with("sts", region_name="us-west-2")


def test_logs(aws_session):
    aws_session._logs = Mock()
    assert aws_session.logs_client
    aws_session.boto_session.client.assert_not_called()
    aws_session._logs = None
    assert aws_session.logs_client
    aws_session.boto_session.client.assert_called_with("logs", region_name="us-west-2")


def test_ecr(aws_session):
    aws_session._ecr = Mock()
    assert aws_session.ecr_client
    aws_session.boto_session.client.assert_not_called()
    aws_session._ecr = None
    assert aws_session.ecr_client
    aws_session.boto_session.client.assert_called_with("ecr", region_name="us-west-2")


@patch("os.path.exists")
@pytest.mark.parametrize(
    "metadata_file_exists, initial_user_agent",
    [
        (True, None),
        (False, None),
        (True, ""),
        (False, ""),
        (True, "Boto3/1.17.18 Python/3.7.10"),
        (False, "Boto3/1.17.18 Python/3.7.10 exec-env/AWS_Lambda_python3.7"),
    ],
)
def test_populates_user_agent(os_path_exists_mock, metadata_file_exists, initial_user_agent):
    request = Mock()
    request.headers = HTTPMessage()
    boto_session = Mock()
    boto_session.region_name = "foobar"
    braket_client = Mock()
    braket_client.meta.region_name = "foobar"
    nbi_metadata_path = "/opt/ml/metadata/resource-metadata.json"
    os_path_exists_mock.return_value = metadata_file_exists
    aws_session = AwsSession(boto_session=boto_session, braket_client=braket_client)
    expected_user_agent = (
        f"BraketSdk/{braket_sdk.__version__} "
        f"BraketSchemas/{braket_schemas.__version__} "
        f"NotebookInstance/{0 if metadata_file_exists else None}"
    )
    os_path_exists_mock.assert_called_with(nbi_metadata_path)

    if initial_user_agent is not None:
        request.headers.add_header("User-Agent", initial_user_agent)
        expected_user_agent = f"{initial_user_agent} {expected_user_agent}"

    aws_session._add_braket_user_agents_handler(request)
    assert request.headers.get("User-Agent") == expected_user_agent


@patch("braket.aws.aws_session.active_trackers")
def test_add_cost_tracker_count(active_trackers_mock, aws_session):
    request = Mock()
    active_trackers_mock.return_value = {"A tracker"}
    aws_session._add_cost_tracker_count_handler(request)
    request.headers.add_header.assert_called_with("Braket-Trackers", "1")
    active_trackers_mock.return_value = {}
    aws_session._add_cost_tracker_count_handler(request)
    request.headers.add_header.assert_called_with("Braket-Trackers", "0")


def test_retrieve_s3_object_body_success(boto_session):
    bucket_name = "braket-integ-test"
    filename = "tasks/test_task_1.json"

    mock_resource = Mock()
    boto_session.resource.return_value = mock_resource
    mock_object = Mock()
    mock_resource.Object.return_value = mock_object
    mock_body_object = Mock()
    mock_object.get.return_value = {"Body": mock_body_object}
    mock_read_object = Mock()
    mock_body_object.read.return_value = mock_read_object
    mock_read_object.decode.return_value = json.dumps(TEST_S3_OBJ_CONTENTS)
    json.dumps(TEST_S3_OBJ_CONTENTS)

    aws_session = AwsSession(boto_session=boto_session)
    return_value = aws_session.retrieve_s3_object_body(bucket_name, filename)
    assert return_value == json.dumps(TEST_S3_OBJ_CONTENTS)
    boto_session.resource.assert_called_with("s3", config=None)

    config = Mock()
    AwsSession(boto_session=boto_session, config=config).retrieve_s3_object_body(
        bucket_name, filename
    )
    boto_session.resource.assert_called_with("s3", config=config)


@pytest.mark.xfail(raises=ClientError)
def test_retrieve_s3_object_body_client_error(boto_session):
    bucket_name = "braket-integ-test"
    filename = "tasks/test_task_1.json"

    mock_resource = Mock()
    boto_session.resource.return_value = mock_resource
    mock_object = Mock()
    mock_resource.Object.return_value = mock_object
    mock_object.get.side_effect = ClientError(
        {"Error": {"Code": "ValidationException", "Message": "NoSuchKey"}}, "Operation"
    )
    aws_session = AwsSession(boto_session=boto_session)
    aws_session.retrieve_s3_object_body(bucket_name, filename)


def test_get_device(boto_session, braket_client):
    return_val = {"deviceArn": "arn1", "deviceName": "name1"}
    braket_client.get_device.return_value = return_val
    aws_session = AwsSession(boto_session=boto_session, braket_client=braket_client)
    metadata = aws_session.get_device("arn1")
    assert return_val == metadata


def test_cancel_quantum_task(aws_session):
    arn = "foo:bar:arn"
    aws_session.braket_client.cancel_quantum_task.return_value = {
        "quantumTaskArn": arn,
        "cancellationStatus": "CANCELLING OR CANCELLED",
    }

    assert aws_session.cancel_quantum_task(arn) is None
    aws_session.braket_client.cancel_quantum_task.assert_called_with(quantumTaskArn=arn)


def test_create_quantum_task(aws_session):
    arn = "foo:bar:arn"
    aws_session.braket_client.create_quantum_task.return_value = {"quantumTaskArn": arn}

    kwargs = {
        "backendArn": "arn:aws:us-west-2:abc:xyz:abc",
        "cwLogGroupArn": "arn:aws:us-west-2:abc:xyz:abc",
        "destinationUrl": "http://s3-us-west-2.amazonaws.com/task-output-bar-1/output.json",
        "program": {"ir": '{"instructions":[]}', "qubitCount": 4},
        "shots": 1,
        "deviceArn": "foo:bar:device_arn",
    }
    assert aws_session.create_quantum_task(**kwargs) == arn
    aws_session.braket_client.create_quantum_task.assert_called_with(**kwargs)


def test_create_quantum_task_with_job_token(aws_session):
    arn = "arn:aws:braket:us-west-2:1234567890:task/task-name"
    job_token = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    aws_session.braket_client.create_quantum_task.return_value = {"quantumTaskArn": arn}

    kwargs = {
        "backendArn": "arn:aws:us-west-2:abc:xyz:abc",
        "cwLogGroupArn": "arn:aws:us-west-2:abc:xyz:abc",
        "destinationUrl": "http://s3-us-west-2.amazonaws.com/task-output-foo-1/output.json",
        "program": {"ir": '{"instructions":[]}', "qubitCount": 4},
        "shots": 1,
        "deviceArn": "foo:bar:device_arn",
    }
    with patch.dict(os.environ, {"AMZN_BRAKET_JOB_TOKEN": job_token}):
        assert aws_session.create_quantum_task(**kwargs) == arn
        kwargs.update({"jobToken": job_token})
        aws_session.braket_client.create_quantum_task.assert_called_with(**kwargs)


def test_get_quantum_task(aws_session):
    arn = "foo:bar:arn"
    status = "STATUS"
    queue_info = ["QueueInfo"]
    return_value = {"quantumTaskArn": arn, "status": status}
    aws_session.braket_client.get_quantum_task.return_value = return_value

    assert aws_session.get_quantum_task(arn) == return_value
    aws_session.braket_client.get_quantum_task.assert_called_with(
        quantumTaskArn=arn, additionalAttributeNames=queue_info
    )


def test_get_quantum_task_retry(aws_session, throttling_response, resource_not_found_response):
    arn = "foo:bar:arn"
    status = "STATUS"
    queue_info = ["QueueInfo"]
    return_value = {"quantumTaskArn": arn, "status": status}

    aws_session.braket_client.get_quantum_task.side_effect = [
        ClientError(resource_not_found_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
        return_value,
    ]

    assert aws_session.get_quantum_task(arn) == return_value
    aws_session.braket_client.get_quantum_task.assert_called_with(
        quantumTaskArn=arn, additionalAttributeNames=queue_info
    )
    assert aws_session.braket_client.get_quantum_task.call_count == 3


def test_get_quantum_task_fail_after_retries(
    aws_session, throttling_response, resource_not_found_response
):
    aws_session.braket_client.get_quantum_task.side_effect = [
        ClientError(resource_not_found_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
    ]

    with pytest.raises(ClientError):
        aws_session.get_quantum_task("some-arn")
    assert aws_session.braket_client.get_quantum_task.call_count == 3


def test_get_quantum_task_does_not_retry_other_exceptions(aws_session):
    exception_response = {
        "Error": {
            "Code": "SomeOtherException",
            "Message": "unit-test-error",
        }
    }

    aws_session.braket_client.get_quantum_task.side_effect = [
        ClientError(exception_response, "unit-test"),
    ]

    with pytest.raises(ClientError):
        aws_session.get_quantum_task("some-arn")
    assert aws_session.braket_client.get_quantum_task.call_count == 1


def test_get_job(aws_session, get_job_response):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    queue_info = ["QueueInfo"]
    aws_session.braket_client.get_job.return_value = get_job_response

    assert aws_session.get_job(arn) == get_job_response
    aws_session.braket_client.get_job.assert_called_with(
        jobArn=arn, additionalAttributeNames=queue_info
    )


def test_get_job_retry(
    aws_session, get_job_response, throttling_response, resource_not_found_response
):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    queue_info = ["QueueInfo"]

    aws_session.braket_client.get_job.side_effect = [
        ClientError(resource_not_found_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
        get_job_response,
    ]

    assert aws_session.get_job(arn) == get_job_response
    aws_session.braket_client.get_job.assert_called_with(
        jobArn=arn, additionalAttributeNames=queue_info
    )
    assert aws_session.braket_client.get_job.call_count == 3


def test_get_job_fail_after_retries(aws_session, throttling_response, resource_not_found_response):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    queue_info = ["QueueInfo"]

    aws_session.braket_client.get_job.side_effect = [
        ClientError(resource_not_found_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
        ClientError(throttling_response, "unit-test"),
    ]

    with pytest.raises(ClientError):
        aws_session.get_job(arn)
    aws_session.braket_client.get_job.assert_called_with(
        jobArn=arn, additionalAttributeNames=queue_info
    )
    assert aws_session.braket_client.get_job.call_count == 3


def test_get_job_does_not_retry_other_exceptions(aws_session):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    queue_info = ["QueueInfo"]
    exception_response = {
        "Error": {
            "Code": "SomeOtherException",
            "Message": "unit-test-error",
        }
    }

    aws_session.braket_client.get_job.side_effect = [
        ClientError(exception_response, "unit-test"),
    ]

    with pytest.raises(ClientError):
        aws_session.get_job(arn)
    aws_session.braket_client.get_job.assert_called_with(
        jobArn=arn, additionalAttributeNames=queue_info
    )
    assert aws_session.braket_client.get_job.call_count == 1


def test_cancel_job(aws_session):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    cancel_job_response = {
        "ResponseMetadata": {
            "RequestId": "857b0893-2073-4ad6-b828-744af8400dfe",
            "HTTPStatusCode": 200,
        },
        "cancellationStatus": "CANCELLING",
        "jobArn": "arn:aws:braket:us-west-2:1234567890:job/job-name",
    }
    aws_session.braket_client.cancel_job.return_value = cancel_job_response

    assert aws_session.cancel_job(arn) == cancel_job_response
    aws_session.braket_client.cancel_job.assert_called_with(jobArn=arn)


@pytest.mark.parametrize(
    "exception_type",
    [
        "ResourceNotFoundException",
        "ValidationException",
        "AccessDeniedException",
        "ThrottlingException",
        "InternalServiceException",
        "ConflictException",
    ],
)
def test_cancel_job_surfaces_errors(exception_type, aws_session):
    arn = "arn:aws:braket:us-west-2:1234567890:job/job-name"
    exception_response = {
        "Error": {
            "Code": "SomeOtherException",
            "Message": "unit-test-error",
        }
    }

    aws_session.braket_client.cancel_job.side_effect = [
        ClientError(exception_response, "unit-test"),
    ]

    with pytest.raises(ClientError):
        aws_session.cancel_job(arn)
    aws_session.braket_client.cancel_job.assert_called_with(jobArn=arn)
    assert aws_session.braket_client.cancel_job.call_count == 1


@pytest.mark.parametrize(
    "input,output",
    [
        (
            {},
            [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn2",
                    "deviceName": "name2",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "OFFLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn3",
                    "deviceName": "name3",
                    "deviceType": "QPU",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname2",
                },
            ],
        ),
        (
            {"names": ["name1"]},
            [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
            ],
        ),
        (
            {"types": ["SIMULATOR"]},
            [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn2",
                    "deviceName": "name2",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "OFFLINE",
                    "providerName": "pname1",
                },
            ],
        ),
        (
            {"statuses": ["ONLINE"]},
            [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn3",
                    "deviceName": "name3",
                    "deviceType": "QPU",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname2",
                },
            ],
        ),
        (
            {"statuses": ["RETIRED"]},
            [
                {
                    "deviceArn": "arn4",
                    "deviceName": "name4",
                    "deviceType": "QPU",
                    "deviceStatus": "RETIRED",
                    "providerName": "pname3",
                },
            ],
        ),
        (
            {"provider_names": ["pname2"]},
            [
                {
                    "deviceArn": "arn3",
                    "deviceName": "name3",
                    "deviceType": "QPU",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname2",
                },
            ],
        ),
        (
            {
                "provider_names": ["pname2"],
                "types": ["QPU"],
                "statuses": ["ONLINE"],
                "names": ["name3"],
            },
            [
                {
                    "deviceArn": "arn3",
                    "deviceName": "name3",
                    "deviceType": "QPU",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname2",
                },
            ],
        ),
        (
            {
                "provider_names": ["pname1"],
                "types": ["SIMULATOR"],
                "statuses": ["ONLINE"],
            },
            [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
            ],
        ),
    ],
)
def test_search_devices(input, output, aws_session):
    return_value = [
        {
            "devices": [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn2",
                    "deviceName": "name2",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "OFFLINE",
                    "providerName": "pname1",
                },
                {
                    "deviceArn": "arn3",
                    "deviceName": "name3",
                    "deviceType": "QPU",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname2",
                },
                {
                    "deviceArn": "arn4",
                    "deviceName": "name4",
                    "deviceType": "QPU",
                    "deviceStatus": "RETIRED",
                    "providerName": "pname3",
                },
            ]
        }
    ]
    mock_paginator = Mock()
    mock_iterator = MagicMock()
    aws_session.braket_client.get_paginator.return_value = mock_paginator
    mock_paginator.paginate.return_value = mock_iterator
    mock_iterator.__iter__.return_value = return_value

    assert aws_session.search_devices(**input) == output


def test_search_devices_arns(aws_session):
    return_value = [
        {
            "devices": [
                {
                    "deviceArn": "arn1",
                    "deviceName": "name1",
                    "deviceType": "SIMULATOR",
                    "deviceStatus": "ONLINE",
                    "providerName": "pname1",
                }
            ]
        }
    ]
    mock_paginator = Mock()
    mock_iterator = MagicMock()
    aws_session.braket_client.get_paginator.return_value = mock_paginator
    mock_paginator.paginate.return_value = mock_iterator
    mock_iterator.__iter__.return_value = return_value

    assert aws_session.search_devices(arns=["arn1"]) == return_value[0]["devices"]
    mock_paginator.paginate.assert_called_with(
        filters=[
            {"name": "deviceArn", "values": ["arn1"]},
        ],
        PaginationConfig={"MaxItems": 100},
    )


def test_create_job(aws_session):
    arn = "foo:bar:arn"
    aws_session.braket_client.create_job.return_value = {"jobArn": arn}

    kwargs = {
        "jobName": "job-name",
        "roleArn": "role-arn",
        "algorithmSpecification": {
            "scriptModeConfig": {
                "entryPoint": "entry-point",
                "s3Uri": "s3-uri",
                "compressionType": "GZIP",
            }
        },
    }
    assert aws_session.create_job(**kwargs) == arn
    aws_session.braket_client.create_job.assert_called_with(**kwargs)


@pytest.mark.parametrize(
    "string, valid",
    (
        ("s3://bucket/key", True),
        ("S3://bucket/key", True),
        ("https://bucket-name-123.s3.us-west-2.amazonaws.com/key/with/dirs", True),
        ("https://bucket-name-123.S3.us-west-2.amazonaws.com/key/with/dirs", True),
        ("https://bucket-name-123.S3.us-west-2.amazonaws.com/", False),
        ("https://bucket-name-123.S3.us-west-2.amazonaws.com", False),
        ("https://S3.us-west-2.amazonaws.com", False),
        ("s3://bucket/", False),
        ("s3://bucket", False),
        ("s3://////", False),
        ("http://bucket/key", False),
        ("bucket/key", False),
    ),
)
def test_is_s3_uri(string, valid):
    assert AwsSession.is_s3_uri(string) == valid


@pytest.mark.parametrize(
    "uri, bucket, key",
    (
        (
            "s3://bucket-name-123/key/with/multiple/dirs",
            "bucket-name-123",
            "key/with/multiple/dirs",
        ),
        (
            "s3://bucket-name-123/key-with_one.dirs",
            "bucket-name-123",
            "key-with_one.dirs",
        ),
        (
            "https://bucket-name-123.s3.us-west-2.amazonaws.com/key/with/dirs",
            "bucket-name-123",
            "key/with/dirs",
        ),
        (
            "https://bucket-name-123.S3.us-west-2.amazonaws.com/key/with/dirs",
            "bucket-name-123",
            "key/with/dirs",
        ),
    ),
)
def test_parse_s3_uri(uri, bucket, key):
    assert bucket, key == AwsSession.parse_s3_uri(uri)


@pytest.mark.parametrize(
    "uri",
    (
        "s3://bucket.name-123/key-with_one.dirs",
        "http://bucket-name-123/key/with/multiple/dirs",
        "bucket-name-123/key/with/multiple/dirs",
        "s3://bucket-name-123/",
        "s3://bucket-name-123",
    ),
)
def test_parse_s3_uri_invalid(uri):
    with pytest.raises(ValueError, match=f"Not a valid S3 uri: {uri}"):
        AwsSession.parse_s3_uri(uri)


@pytest.mark.parametrize(
    "bucket, dirs",
    [
        ("bucket", ("d1", "d2", "d3")),
        ("bucket-123-braket", ("dir",)),
        pytest.param(
            "braket",
            (),
            marks=pytest.mark.xfail(raises=ValueError, strict=True),
        ),
    ],
)
def test_construct_s3_uri(bucket, dirs):
    parsed_bucket, parsed_key = AwsSession.parse_s3_uri(AwsSession.construct_s3_uri(bucket, *dirs))
    assert parsed_bucket == bucket
    assert parsed_key == "/".join(dirs)


def test_get_default_jobs_role(aws_session, job_role_arn, job_role_name):
    iam_client = boto3.client("iam")
    with Stubber(iam_client) as stub:
        stub.add_response(
            "list_roles",
            {
                "Roles": [
                    {
                        "Arn": "arn:aws:iam::0000000000:role/nonJobsRole",
                        "RoleName": "nonJobsRole",
                        "Path": "/",
                        "RoleId": "nonJobsRole-213453451345-431513",
                        "CreateDate": time.time(),
                    }
                ]
                * 100,
                "IsTruncated": True,
                "Marker": "resp-marker",
            },
            {"PathPrefix": "/service-role/"},
        )
        stub.add_response(
            "list_roles",
            {
                "Roles": [
                    {
                        "Arn": job_role_arn,
                        "RoleName": job_role_name,
                        "Path": "/",
                        "RoleId": f"{job_role_name}-213453451345-431513",
                        "CreateDate": time.time(),
                    }
                ],
                "IsTruncated": False,
            },
            {"Marker": "resp-marker", "PathPrefix": "/service-role/"},
        )
        aws_session._iam = iam_client
        assert aws_session.get_default_jobs_role() == job_role_arn


def test_get_default_jobs_role_not_found(aws_session, job_role_arn, job_role_name):
    iam_client = boto3.client("iam")
    with Stubber(iam_client) as stub:
        stub.add_response(
            "list_roles",
            {
                "Roles": [
                    {
                        "Arn": "arn:aws:iam::0000000000:role/nonJobsRole",
                        "RoleName": "nonJobsRole",
                        "Path": "/",
                        "RoleId": "nonJobsRole-213453451345-431513",
                        "CreateDate": time.time(),
                    }
                ]
                * 100,
                "IsTruncated": True,
                "Marker": "resp-marker",
            },
            {"PathPrefix": "/service-role/"},
        )
        stub.add_response(
            "list_roles",
            {
                "Roles": [
                    {
                        "Arn": "arn:aws:iam::0000000000:role/nonJobsRole2",
                        "RoleName": "nonJobsRole2",
                        "Path": "/",
                        "RoleId": "nonJobsRole2-213453451345-431513",
                        "CreateDate": time.time(),
                    }
                ],
                "IsTruncated": False,
            },
            {"Marker": "resp-marker", "PathPrefix": "/service-role/"},
        )
        aws_session._iam = iam_client
        with pytest.raises(RuntimeError):
            aws_session.get_default_jobs_role()


def test_upload_to_s3(aws_session):
    filename = "file.txt"
    s3_uri = "s3://bucket-123/key"
    bucket, key = "bucket-123", "key"
    aws_session.upload_to_s3(filename, s3_uri)
    aws_session._s3.upload_file.assert_called_with(filename, bucket, key)


def test_account_id_idempotency(aws_session, account_id):
    acc_id = aws_session.account_id
    assert acc_id == aws_session.account_id
    assert acc_id == account_id


def test_upload_local_data(aws_session):
    with tempfile.TemporaryDirectory() as temp_dir:
        os.chdir(temp_dir)

        Path("input-dir", "pref-dir", "sub-pref-dir").mkdir(parents=True)
        Path("input-dir", "not-pref-dir").mkdir()

        # these should all get uploaded
        Path("input-dir", "pref-dir", "sub-pref-dir", "very-nested.txt").touch()
        Path("input-dir", "pref-dir", "nested.txt").touch()
        Path("input-dir", "pref.txt").touch()
        Path("input-dir", "pref-and-more.txt").touch()

        # these should not
        Path("input-dir", "false-pref.txt").touch()
        Path("input-dir", "not-pref-dir", "pref-fake.txt").touch()

        aws_session.upload_to_s3 = Mock()
        aws_session.upload_local_data("input-dir/pref", "s3://bucket/pref")
        call_args = {args for args, kwargs in aws_session.upload_to_s3.call_args_list}
        assert call_args == {
            (
                str(Path("input-dir", "pref-dir", "sub-pref-dir", "very-nested.txt")),
                "s3://bucket/pref-dir/sub-pref-dir/very-nested.txt",
            ),
            (str(Path("input-dir", "pref-dir", "nested.txt")), "s3://bucket/pref-dir/nested.txt"),
            (str(Path("input-dir", "pref.txt")), "s3://bucket/pref.txt"),
            (str(Path("input-dir", "pref-and-more.txt")), "s3://bucket/pref-and-more.txt"),
        }
        os.chdir("..")


def test_upload_local_data_absolute(aws_session):
    with tempfile.TemporaryDirectory() as temp_dir:
        Path(temp_dir, "input-dir", "pref-dir", "sub-pref-dir").mkdir(parents=True)
        Path(temp_dir, "input-dir", "not-pref-dir").mkdir()

        # these should all get uploaded
        Path(temp_dir, "input-dir", "pref-dir", "sub-pref-dir", "very-nested.txt").touch()
        Path(temp_dir, "input-dir", "pref-dir", "nested.txt").touch()
        Path(temp_dir, "input-dir", "pref.txt").touch()
        Path(temp_dir, "input-dir", "pref-and-more.txt").touch()

        # these should not
        Path(temp_dir, "input-dir", "false-pref.txt").touch()
        Path(temp_dir, "input-dir", "not-pref-dir", "pref-fake.txt").touch()

        aws_session.upload_to_s3 = Mock()
        aws_session.upload_local_data(str(Path(temp_dir, "input-dir", "pref")), "s3://bucket/pref")
        call_args = {args for args, kwargs in aws_session.upload_to_s3.call_args_list}
        assert call_args == {
            (
                str(Path(temp_dir, "input-dir", "pref-dir", "sub-pref-dir", "very-nested.txt")),
                "s3://bucket/pref-dir/sub-pref-dir/very-nested.txt",
            ),
            (
                str(Path(temp_dir, "input-dir", "pref-dir", "nested.txt")),
                "s3://bucket/pref-dir/nested.txt",
            ),
            (str(Path(temp_dir, "input-dir", "pref.txt")), "s3://bucket/pref.txt"),
            (
                str(Path(temp_dir, "input-dir", "pref-and-more.txt")),
                "s3://bucket/pref-and-more.txt",
            ),
        }


def test_download_from_s3(aws_session):
    filename = "model.tar.gz"
    s3_uri = (
        "s3://amazon-braket-jobs/job-path/output/"
        "BraketJob-875981177017-job-test-20210628140446/output/model.tar.gz"
    )
    bucket, key = (
        "amazon-braket-jobs",
        "job-path/output/BraketJob-875981177017-job-test-20210628140446/output/model.tar.gz",
    )
    aws_session.download_from_s3(s3_uri, filename)
    aws_session._s3.download_file.assert_called_with(bucket, key, filename)


def test_copy_identical_s3(aws_session):
    s3_uri = "s3://bucket/key"
    aws_session.copy_s3_object(s3_uri, s3_uri)
    aws_session.boto_session.client.return_value.copy.assert_not_called()


def test_copy_s3(aws_session):
    source_s3_uri = "s3://here/now"
    dest_s3_uri = "s3://there/then"
    source_bucket, source_key = AwsSession.parse_s3_uri(source_s3_uri)
    dest_bucket, dest_key = AwsSession.parse_s3_uri(dest_s3_uri)
    aws_session.copy_s3_object(source_s3_uri, dest_s3_uri)
    aws_session._s3.copy.assert_called_with(
        {
            "Bucket": source_bucket,
            "Key": source_key,
        },
        dest_bucket,
        dest_key,
    )


def test_copy_identical_s3_directory(aws_session):
    s3_uri = "s3://bucket/prefix/"
    aws_session.copy_s3_directory(s3_uri, s3_uri)
    aws_session.boto_session.client.return_value.copy.assert_not_called()


def test_copy_s3_directory(aws_session):
    aws_session.list_keys = Mock(return_value=[f"now/key-{i}" for i in range(5)])
    source_s3_uri = "s3://here/now"
    dest_s3_uri = "s3://there/then"
    aws_session.copy_s3_directory(source_s3_uri, dest_s3_uri)
    for i in range(5):
        aws_session.s3_client.copy.assert_any_call(
            {
                "Bucket": "here",
                "Key": f"now/key-{i}",
            },
            "there",
            f"then/key-{i}",
        )


def test_list_keys(aws_session):
    bucket, prefix = "bucket", "prefix"
    aws_session.s3_client.list_objects_v2.side_effect = [
        {
            "IsTruncated": True,
            "Contents": [
                {"Key": "copy-test/copy.txt"},
                {"Key": "copy-test/copy2.txt"},
            ],
            "NextContinuationToken": "next-continuation-token",
        },
        {
            "IsTruncated": False,
            "Contents": [
                {"Key": "copy-test/nested/double-nested/double-nested.txt"},
                {"Key": "copy-test/nested/nested.txt"},
            ],
        },
    ]
    keys = aws_session.list_keys(bucket, prefix)
    assert keys == [
        "copy-test/copy.txt",
        "copy-test/copy2.txt",
        "copy-test/nested/double-nested/double-nested.txt",
        "copy-test/nested/nested.txt",
    ]


def test_default_bucket(aws_session, account_id):
    region = "test-region-0"
    aws_session.boto_session.region_name = region
    assert aws_session.default_bucket() == f"amazon-braket-{region}-{account_id}"


def test_default_bucket_given(aws_session):
    default_bucket = "default_bucket"
    aws_session._default_bucket = default_bucket
    assert aws_session.default_bucket() == default_bucket
    aws_session._s3.create_bucket.assert_not_called()


@patch.dict("os.environ", {"AMZN_BRAKET_OUT_S3_BUCKET": "default_bucket_env"})
def test_default_bucket_env_variable(boto_session, braket_client):
    aws_session = AwsSession(boto_session=boto_session, braket_client=braket_client)
    assert aws_session.default_bucket() == "default_bucket_env"


@pytest.mark.parametrize(
    "region",
    (
        "test-region-0",
        "us-east-1",
    ),
)
def test_create_s3_bucket_if_it_does_not_exist(aws_session, region, account_id):
    bucket = f"amazon-braket-{region}-{account_id}"
    aws_session._create_s3_bucket_if_it_does_not_exist(bucket, region)
    kwargs = {
        "Bucket": bucket,
        "CreateBucketConfiguration": {
            "LocationConstraint": region,
        },
    }
    if region == "us-east-1":
        del kwargs["CreateBucketConfiguration"]
    aws_session._s3.create_bucket.assert_called_with(**kwargs)
    aws_session._s3.put_public_access_block.assert_called_with(
        Bucket=bucket,
        PublicAccessBlockConfiguration={
            "BlockPublicAcls": True,
            "IgnorePublicAcls": True,
            "BlockPublicPolicy": True,
            "RestrictPublicBuckets": True,
        },
    )


@pytest.mark.parametrize(
    "error",
    (
        ClientError(
            {
                "Error": {
                    "Code": "BucketAlreadyOwnedByYou",
                    "Message": "Your previous request to create the named bucket succeeded "
                    "and you already own it.",
                }
            },
            "CreateBucket",
        ),
        ClientError(
            {
                "Error": {
                    "Code": "OperationAborted",
                    "Message": "A conflicting conditional operation is currently in progress "
                    "against this resource. Please try again.",
                }
            },
            "CreateBucket",
        ),
        pytest.param(
            ClientError(
                {
                    "Error": {
                        "Code": "OtherCode",
                        "Message": "This should fail properly.",
                    }
                },
                "CreateBucket",
            ),
            marks=pytest.mark.xfail(raises=ClientError, strict=True),
        ),
    ),
)
def test_create_s3_bucket_if_it_does_not_exist_error(aws_session, error, account_id):
    region = "test-region-0"
    bucket = f"amazon-braket-{region}-{account_id}"
    aws_session._s3.create_bucket.side_effect = error
    aws_session._create_s3_bucket_if_it_does_not_exist(bucket, region)


@pytest.mark.xfail(raises=ValueError)
def test_bucket_already_exists_for_another_account(aws_session):
    exception_response = {
        "Error": {
            "Code": "BucketAlreadyExists",
            "Message": "This should fail properly.",
        }
    }
    bucket_name, region = "some-bucket-123", "test-region"
    aws_session._s3.create_bucket.side_effect = ClientError(exception_response, "CreateBucket")
    aws_session._create_s3_bucket_if_it_does_not_exist(bucket_name, region)


@pytest.mark.parametrize(
    "limit, next_token",
    (
        (None, None),
        (10, None),
        (None, "next-token"),
        (10, "next-token"),
    ),
)
def test_describe_log_streams(aws_session, limit, next_token):
    aws_session._logs = Mock()

    log_group = "log_group"
    log_stream_prefix = "log_stream_prefix"

    describe_log_stream_args = {
        "logGroupName": log_group,
        "logStreamNamePrefix": log_stream_prefix,
        "orderBy": "LogStreamName",
    }

    if limit:
        describe_log_stream_args.update({"limit": limit})

    if next_token:
        describe_log_stream_args.update({"nextToken": next_token})

    aws_session.describe_log_streams(log_group, log_stream_prefix, limit, next_token)

    aws_session._logs.describe_log_streams.assert_called_with(**describe_log_stream_args)


@pytest.mark.parametrize(
    "next_token",
    (None, "next-token"),
)
def test_get_log_events(aws_session, next_token):
    aws_session._logs = Mock()

    log_group = "log_group"
    log_stream_name = "log_stream_name"
    start_time = "timestamp"
    start_from_head = True

    log_events_args = {
        "logGroupName": log_group,
        "logStreamName": log_stream_name,
        "startTime": start_time,
        "startFromHead": start_from_head,
    }

    if next_token:
        log_events_args.update({"nextToken": next_token})

    aws_session.get_log_events(log_group, log_stream_name, start_time, start_from_head, next_token)

    aws_session._logs.get_log_events.assert_called_with(**log_events_args)


@patch("boto3.Session")
def test_copy_session(boto_session_init, aws_session):
    boto_session_init.return_value = Mock()
    aws_session._braket_user_agents = "foo/bar"
    copied_session = AwsSession.copy_session(aws_session, "us-west-2")
    boto_session_init.assert_called_with(
        region_name="us-west-2",
        profile_name="test-profile",
    )
    assert copied_session._braket_user_agents == "foo/bar"
    assert copied_session._default_bucket is None


@patch("boto3.Session")
def test_copy_explicit_session(boto_session_init, aws_explicit_session):
    boto_session_init.return_value = Mock()
    AwsSession.copy_session(aws_explicit_session, "us-west-2")
    boto_session_init.assert_called_with(
        aws_access_key_id="access key",
        aws_secret_access_key="secret key",
        aws_session_token="token",
        region_name="us-west-2",
        profile_name="test-profile",
    )


@patch("boto3.Session")
def test_copy_env_session(boto_session_init, aws_env_session):
    boto_session_init.return_value = Mock()
    AwsSession.copy_session(aws_env_session, "us-west-2")
    boto_session_init.assert_called_with(
        region_name="us-west-2",
    )


@patch("boto3.Session")
def test_copy_session_custom_default_bucket(mock_boto, aws_session):
    mock_boto.return_value.region_name = "us-test-1"
    aws_session._default_bucket = "my-own-default"
    aws_session._custom_default_bucket = True
    copied_session = AwsSession.copy_session(aws_session)
    assert copied_session._default_bucket == "my-own-default"


def test_add_braket_user_agent(aws_session):
    user_agent = "newAgent/1.0"
    aws_session.add_braket_user_agent(user_agent)
    aws_session.add_braket_user_agent(user_agent)
    aws_session._braket_user_agents.count(user_agent) == 1


def test_get_full_image_tag(aws_session):
    aws_session.ecr_client.batch_get_image.side_effect = (
        {"images": [{"imageId": {"imageDigest": "my-digest"}}]},
        {
            "images": [
                {"imageId": {"imageTag": "my-tag"}},
                {"imageId": {"imageTag": "my-tag-py3"}},
                {"imageId": {"imageTag": "my-tag-py310"}},
                {"imageId": {"imageTag": "latest"}},
            ]
        },
        AssertionError("Image tag not cached"),
    )
    image_uri = "123456.image_uri/repo-name:my-tag"
    assert aws_session.get_full_image_tag(image_uri) == "my-tag-py310"
    assert aws_session.get_full_image_tag(image_uri) == "my-tag-py310"


def test_get_full_image_tag_no_py_info(aws_session):
    aws_session.ecr_client.batch_get_image.side_effect = (
        {"images": [{"imageId": {"imageDigest": "my-digest"}}]},
        {
            "images": [
                {"imageId": {"imageTag": "my-tag"}},
                {"imageId": {"imageTag": "latest"}},
            ]
        },
    )
    image_uri = "123456.image_uri/repo-name:my-tag"

    no_py_info = "Full image tag missing."
    with pytest.raises(ValueError, match=no_py_info):
        aws_session.get_full_image_tag(image_uri)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest
from gate_model_device_testing_utils import (
    no_result_types_bell_pair_testing,
    noisy_circuit_1qubit_noise_full_probability,
    noisy_circuit_2qubit_noise_full_probability,
    qubit_ordering_testing,
    result_types_all_selected_testing,
    result_types_bell_pair_full_probability_testing,
    result_types_bell_pair_marginal_probability_testing,
    result_types_hermitian_testing,
    result_types_noncommuting_all,
    result_types_noncommuting_flipped_targets_testing,
    result_types_noncommuting_testing,
    result_types_nonzero_shots_bell_pair_testing,
    result_types_tensor_hermitian_hermitian_testing,
    result_types_tensor_x_y_testing,
    result_types_tensor_y_hermitian_testing,
    result_types_tensor_z_h_y_testing,
    result_types_tensor_z_hermitian_testing,
    result_types_tensor_z_z_testing,
)

from braket.devices import LocalSimulator

DEVICE = LocalSimulator("braket_dm")
SHOTS = 8000


def test_no_result_types_bell_pair(caplog):
    no_result_types_bell_pair_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


def test_qubit_ordering(caplog):
    qubit_ordering_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


def test_result_types_nonzero_shots_bell_pair(caplog):
    result_types_nonzero_shots_bell_pair_testing(DEVICE, {"shots": SHOTS})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_bell_pair_full_probability(shots, caplog):
    result_types_bell_pair_full_probability_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_bell_pair_marginal_probability(shots, caplog):
    result_types_bell_pair_marginal_probability_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_hermitian(shots, caplog):
    result_types_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_x_y(shots, caplog):
    result_types_tensor_x_y_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_z(shots, caplog):
    result_types_tensor_z_z_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_h_y(shots, caplog):
    result_types_tensor_z_h_y_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_z_hermitian(shots, caplog):
    result_types_tensor_z_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_hermitian_hermitian(shots, caplog):
    result_types_tensor_hermitian_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_tensor_y_hermitian(shots, caplog):
    result_types_tensor_y_hermitian_testing(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_result_types_all_selected(shots, caplog):
    result_types_all_selected_testing(DEVICE, {"shots": shots})
    assert not caplog.text


def test_result_types_noncommuting(caplog):
    result_types_noncommuting_testing(DEVICE, {})
    assert not caplog.text


def test_result_types_noncommuting_flipped_targets(caplog):
    result_types_noncommuting_flipped_targets_testing(DEVICE, {})
    assert not caplog.text


def test_result_types_noncommuting_all(caplog):
    result_types_noncommuting_all(DEVICE, {})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_noisy_circuit_1qubit_noise(shots, caplog):
    noisy_circuit_1qubit_noise_full_probability(DEVICE, {"shots": shots})
    assert not caplog.text


@pytest.mark.parametrize("shots", [0, SHOTS])
def test_noisy_circuit_2qubit_noise(shots, caplog):
    noisy_circuit_2qubit_noise_full_probability(DEVICE, {"shots": shots})
    assert not caplog.text



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import numbers
from collections.abc import Sequence
from copy import deepcopy
from typing import Union

import numpy as np

from braket.circuits.gate import Gate
from braket.circuits.quantum_operator import QuantumOperator
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    SerializationProperties,
)
from braket.registers.qubit_set import QubitSet


class Observable(QuantumOperator):
    """Class `Observable` to represent a quantum observable.

    Objects of this type can be used as input to `ResultType.Sample`, `ResultType.Variance`,
    `ResultType.Expectation` to specify the measurement basis.
    """

    def __init__(self, qubit_count: int, ascii_symbols: Sequence[str]):
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)
        self._coef = 1

    def _unscaled(self) -> Observable:
        return Observable(qubit_count=self.qubit_count, ascii_symbols=self.ascii_symbols)

    def to_ir(
        self,
        target: QubitSet | None = None,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: SerializationProperties | None = None,
    ) -> Union[str, list[Union[str, list[list[list[float]]]]]]:
        """Returns the IR representation for the observable

        Args:
            target (QubitSet | None): target qubit(s). Defaults to None.
            ir_type(IRType) : The IRType to use for converting the result type object to its
                IR representation. Defaults to IRType.JAQCD.
            serialization_properties (SerializationProperties | None): The serialization properties
                to use while serializing the object to the IR representation. The serialization
                properties supplied must correspond to the supplied `ir_type`. Defaults to None.

        Returns:
            Union[str, list[Union[str, list[list[list[float]]]]]]: The IR representation for
            the observable.

        Raises:
            ValueError: If the supplied `ir_type` is not supported, or if the supplied serialization
                properties don't correspond to the `ir_type`.
        """
        if ir_type == IRType.JAQCD:
            return self._to_jaqcd()
        elif ir_type == IRType.OPENQASM:
            if serialization_properties and not isinstance(
                serialization_properties, OpenQASMSerializationProperties
            ):
                raise ValueError(
                    "serialization_properties must be of type OpenQASMSerializationProperties "
                    "for IRType.OPENQASM."
                )
            return self._to_openqasm(
                serialization_properties or OpenQASMSerializationProperties(), target
            )
        else:
            raise ValueError(f"Supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self) -> list[Union[str, list[list[list[float]]]]]:
        """Returns the JAQCD representation of the observable."""
        raise NotImplementedError("to_jaqcd has not been implemented yet.")

    def _to_openqasm(
        self,
        serialization_properties: OpenQASMSerializationProperties,
        target: QubitSet | None = None,
    ) -> str:
        """Returns the openqasm string representation of the result type.

        Args:
            serialization_properties (OpenQASMSerializationProperties): The serialization properties
                to use while serializing the object to the IR representation.
            target (QubitSet | None): target qubit(s). Defaults to None.

        Returns:
            str: Representing the openqasm representation of the result type.
        """
        raise NotImplementedError("to_openqasm has not been implemented yet.")

    @property
    def coefficient(self) -> int:
        """The coefficient of the observable.

        Returns:
            int: coefficient value of the observable.
        """
        return self._coef

    @property
    def basis_rotation_gates(self) -> tuple[Gate, ...]:
        """Returns the basis rotation gates for this observable.

        Returns:
            tuple[Gate, ...]: The basis rotation gates for this observable.
        """
        raise NotImplementedError

    @property
    def eigenvalues(self) -> np.ndarray:
        """Returns the eigenvalues of this observable.

        Returns:
            np.ndarray: The eigenvalues of this observable.
        """
        raise NotImplementedError

    def eigenvalue(self, index: int) -> float:
        """Returns the eigenvalue of this observable at the given index.

        The eigenvalues are ordered by their corresponding computational basis state
        after diagonalization.

        Args:
            index (int): The index of the desired eigenvalue

        Returns:
            float: The `index` th eigenvalue of the observable.
        """
        raise NotImplementedError

    @classmethod
    def register_observable(cls, observable: Observable) -> None:
        """Register an observable implementation by adding it into the `Observable` class.

        Args:
            observable (Observable): Observable class to register.
        """
        setattr(cls, observable.__name__, observable)

    def __matmul__(self, other: Observable) -> Observable.TensorProduct:
        if isinstance(other, Observable):
            return Observable.TensorProduct([self, other])

        raise ValueError("Can only perform tensor products between observables.")

    def __mul__(self, other: Observable) -> Observable:
        """Scalar multiplication"""
        if isinstance(other, numbers.Number):
            observable_copy = deepcopy(self)
            observable_copy._coef *= other
            return observable_copy
        raise TypeError("Observable coefficients must be numbers.")

    def __rmul__(self, other: Observable) -> Observable:
        return self * other

    def __add__(self, other: Observable):
        if not isinstance(other, Observable):
            raise ValueError("Can only perform addition between observables.")

        return Observable.Sum([self, other])

    def __sub__(self, other: Observable):
        if not isinstance(other, Observable):
            raise ValueError("Can only perform subtraction between observables.")

        return self + (-1 * other)

    def __repr__(self) -> str:
        return f"{self.name}('qubit_count': {self.qubit_count})"

    def __eq__(self, other: Observable) -> bool:
        if isinstance(other, Observable):
            return self.name == other.name
        return NotImplemented


class StandardObservable(Observable):
    """Class `StandardObservable` to represent a Pauli-like quantum observable with
    eigenvalues of (+1, -1).
    """

    def __init__(self, ascii_symbols: Sequence[str]):
        super().__init__(qubit_count=1, ascii_symbols=ascii_symbols)
        self._eigenvalues = (1.0, -1.0)  # immutable

    def _unscaled(self) -> StandardObservable:
        return StandardObservable(ascii_symbols=self.ascii_symbols)

    @property
    def eigenvalues(self) -> np.ndarray:
        return self.coefficient * np.array(self._eigenvalues)

    def eigenvalue(self, index: int) -> float:
        return self.coefficient * self._eigenvalues[index]

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        return tuple(
            f"{self.coefficient if self.coefficient != 1 else ''}{ascii_symbol}"
            for ascii_symbol in self._ascii_symbols
        )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import io
from unittest.mock import patch

import pytest

from braket.tracking.pricing import Pricing


@pytest.fixture
def mock_http():
    with patch("urllib3.PoolManager") as http_mock:
        http_mock().request.return_value = io.BytesIO(
            b"""line1
line2
line3
line4
line5
A,B
1,1
1,2
"""
        )
        yield http_mock()


def test_search_prices(mock_http):
    pricer = Pricing()
    assert pricer.price_search(A="0") == []
    assert pricer.price_search(A="1", B="1") == [{"A": "1", "B": "1"}]
    assert pricer.price_search(A="1") == [{"A": "1", "B": "1"}, {"A": "1", "B": "2"}]


@patch.dict("os.environ", {"BRAKET_PRICE_OFFERS_URL": "https://myurl"})
def test_price_offer_env_var(mock_http):
    pricer = Pricing()
    pricer.get_prices()

    mock_http.request.assert_called_with("GET", "https://myurl", preload_content=False)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from pathlib import Path
from typing import Any

from braket.jobs.environment_variables import get_checkpoint_dir, get_job_name, get_results_dir
from braket.jobs.serialization import deserialize_values, serialize_values
from braket.jobs_data import PersistedJobData, PersistedJobDataFormat


def save_job_checkpoint(
    checkpoint_data: dict[str, Any],
    checkpoint_file_suffix: str = "",
    data_format: PersistedJobDataFormat = PersistedJobDataFormat.PLAINTEXT,
) -> None:
    """Saves the specified `checkpoint_data` to the local output directory, specified by
    the container environment variable `CHECKPOINT_DIR`, with the filename
    `f"{job_name}(_{checkpoint_file_suffix}).json"`. The `job_name` refers to the name of the
    current job and is retrieved from the container environment variable `JOB_NAME`. The
    `checkpoint_data` values are serialized to the specified `data_format`.

    Note: This function for storing the checkpoints is only for use inside the job container
          as it writes data to directories and references env variables set in the containers.


    Args:
        checkpoint_data (dict[str, Any]): Dict that specifies the checkpoint data to be persisted.
        checkpoint_file_suffix (str): str that specifies the file suffix to be used for
            the checkpoint filename. The resulting filename
            `f"{job_name}(_{checkpoint_file_suffix}).json"` is used to save the checkpoints.
            Default: ""
        data_format (PersistedJobDataFormat): The data format used to serialize the
            values. Note that for `PICKLED` data formats, the values are base64 encoded
            after serialization. Default: PersistedJobDataFormat.PLAINTEXT

    Raises:
        ValueError: If the supplied `checkpoint_data` is `None` or empty.
    """
    if not checkpoint_data:
        raise ValueError("The checkpoint_data argument cannot be empty.")
    checkpoint_directory = get_checkpoint_dir()
    job_name = get_job_name()
    checkpoint_file_path = (
        f"{checkpoint_directory}/{job_name}_{checkpoint_file_suffix}.json"
        if checkpoint_file_suffix
        else f"{checkpoint_directory}/{job_name}.json"
    )
    with open(checkpoint_file_path, "w") as f:
        serialized_data = serialize_values(checkpoint_data or {}, data_format)
        persisted_data = PersistedJobData(dataDictionary=serialized_data, dataFormat=data_format)
        f.write(persisted_data.json())


def load_job_checkpoint(
    job_name: str | None = None, checkpoint_file_suffix: str = ""
) -> dict[str, Any]:
    """Loads the job checkpoint data stored for the job named 'job_name', with the checkpoint
    file that ends with the `checkpoint_file_suffix`. The `job_name` can refer to any job whose
    checkpoint data you expect to be available in the file path specified by the `CHECKPOINT_DIR`
    container environment variable. If not provided, this function will use the currently running
    job's name.

    Note: This function for loading hybrid job checkpoints is only for use inside the job container
          as it writes data to directories and references env variables set in the containers.


    Args:
        job_name (str | None): str that specifies the name of the job whose checkpoints
            are to be loaded. Default: current job name.

        checkpoint_file_suffix (str): str specifying the file suffix that is used to
            locate the checkpoint file to load. The resulting file name
            `f"{job_name}(_{checkpoint_file_suffix}).json"` is used to locate the
            checkpoint file. Default: ""

    Returns:
        dict[str, Any]: Dict that contains the checkpoint data persisted in the checkpoint file.

    Raises:
        FileNotFoundError: If the file `f"{job_name}(_{checkpoint_file_suffix})"` could not be found
            in the directory specified by the container environment variable `CHECKPOINT_DIR`.
        ValueError: If the data stored in the checkpoint file can't be deserialized (possibly due to
            corruption).
    """
    job_name = job_name or get_job_name()
    checkpoint_directory = get_checkpoint_dir()
    checkpoint_file_path = (
        f"{checkpoint_directory}/{job_name}_{checkpoint_file_suffix}.json"
        if checkpoint_file_suffix
        else f"{checkpoint_directory}/{job_name}.json"
    )
    with open(checkpoint_file_path) as f:
        persisted_data = PersistedJobData.parse_raw(f.read())
        deserialized_data = deserialize_values(
            persisted_data.dataDictionary, persisted_data.dataFormat
        )
        return deserialized_data


def _load_persisted_data(filename: str | Path | None = None) -> PersistedJobData:
    filename = filename or Path(get_results_dir()) / "results.json"
    try:
        with open(filename) as f:
            return PersistedJobData.parse_raw(f.read())
    except FileNotFoundError:
        return PersistedJobData(
            dataDictionary={},
            dataFormat=PersistedJobDataFormat.PLAINTEXT,
        )


def load_job_result(filename: str | Path | None = None) -> dict[str, Any]:
    """Loads job result of currently running job.

    Args:
        filename (str | Path | None): Location of job results. Default `results.json` in job
            results directory in a job instance or in working directory locally. This file
            must be in the format used by `save_job_result`.

    Returns:
        dict[str, Any]: Job result data of current job
    """
    persisted_data = _load_persisted_data(filename)
    deserialized_data = deserialize_values(persisted_data.dataDictionary, persisted_data.dataFormat)
    return deserialized_data


def save_job_result(
    result_data: dict[str, Any] | Any,
    data_format: PersistedJobDataFormat | None = None,
) -> None:
    """Saves the `result_data` to the local output directory that is specified by the container
    environment variable `AMZN_BRAKET_JOB_RESULTS_DIR`, with the filename 'results.json'.
    The `result_data` values are serialized to the specified `data_format`.

    Note: This function for storing the results is only for use inside the job container
          as it writes data to directories and references env variables set in the containers.

    Args:
        result_data (dict[str, Any] | Any): Dict that specifies the result data to be
            persisted. If result data is not a dict, then it will be wrapped as
            `{"result": result_data}`.
        data_format (PersistedJobDataFormat | None): The data format used to serialize the
            values. Note that for `PICKLED` data formats, the values are base64 encoded
            after serialization. Default: PersistedJobDataFormat.PLAINTEXT.

    Raises:
        TypeError: Unsupported data format.
    """
    if not isinstance(result_data, dict):
        result_data = {"result": result_data}

    current_persisted_data = _load_persisted_data()

    if current_persisted_data.dataFormat == PersistedJobDataFormat.PICKLED_V4:
        # if results are already pickled, maintain pickled format
        # if user explicitly specifies plaintext, raise error
        if data_format == PersistedJobDataFormat.PLAINTEXT:
            raise TypeError(
                "Cannot update results object serialized with "
                f"{current_persisted_data.dataFormat.value} using data format "
                f"{data_format.value}."
            )

        data_format = PersistedJobDataFormat.PICKLED_V4

    # if not specified or already pickled, default to plaintext
    data_format = data_format or PersistedJobDataFormat.PLAINTEXT

    current_results = deserialize_values(
        current_persisted_data.dataDictionary,
        current_persisted_data.dataFormat,
    )
    updated_results = current_results | result_data

    with open(Path(get_results_dir()) / "results.json", "w") as f:
        serialized_data = serialize_values(updated_results or {}, data_format)
        persisted_data = PersistedJobData(dataDictionary=serialized_data, dataFormat=data_format)
        f.write(persisted_data.json())



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json
import os
import tempfile
from dataclasses import dataclass
from unittest.mock import patch

import numpy as np
import pytest

from braket.jobs.data_persistence import (
    load_job_checkpoint,
    load_job_result,
    save_job_checkpoint,
    save_job_result,
)
from braket.jobs_data import PersistedJobDataFormat


@pytest.mark.parametrize(
    "job_name, file_suffix, data_format, checkpoint_data, expected_saved_data",
    [
        (
            "job_plaintext_simple_dict",
            "",
            PersistedJobDataFormat.PLAINTEXT,
            {"converged": True, "energy": -0.2},
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {"converged": True, "energy": -0.2},
                    "dataFormat": "plaintext",
                }
            ),
        ),
        (
            "job_pickled_simple_dict",
            "suffix1",
            PersistedJobDataFormat.PICKLED_V4,
            {"converged": True, "energy": -0.2},
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {
                        "converged": "gASILg==\n",
                        "energy": "gASVCgAAAAAAAABHv8mZmZmZmZou\n",
                    },
                    "dataFormat": "pickled_v4",
                }
            ),
        ),
    ],
)
def test_save_job_checkpoint(
    job_name, file_suffix, data_format, checkpoint_data, expected_saved_data
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            save_job_checkpoint(checkpoint_data, file_suffix, data_format)

        expected_file_location = (
            f"{tmp_dir}/{job_name}_{file_suffix}.json"
            if file_suffix
            else f"{tmp_dir}/{job_name}.json"
        )
        with open(expected_file_location, "r") as expected_file:
            assert expected_file.read() == expected_saved_data


@pytest.mark.xfail(raises=ValueError)
@pytest.mark.parametrize("checkpoint_data", [{}, None])
def test_save_job_checkpoint_raises_error_empty_data(checkpoint_data):
    job_name = "foo"
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            save_job_checkpoint(checkpoint_data)


@pytest.mark.parametrize(
    "job_name, file_suffix, data_format, saved_data, expected_checkpoint_data",
    [
        (
            "job_plaintext_simple_dict",
            "",
            PersistedJobDataFormat.PLAINTEXT,
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {"converged": True, "energy": -0.2},
                    "dataFormat": "plaintext",
                }
            ),
            {"converged": True, "energy": -0.2},
        ),
        (
            "job_pickled_simple_dict",
            "",
            PersistedJobDataFormat.PICKLED_V4,
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {
                        "converged": "gASILg==\n",
                        "energy": "gASVCgAAAAAAAABHv8mZmZmZmZou\n",
                    },
                    "dataFormat": "pickled_v4",
                }
            ),
            {"converged": True, "energy": -0.2},
        ),
    ],
)
def test_load_job_checkpoint(
    job_name, file_suffix, data_format, saved_data, expected_checkpoint_data
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        file_path = (
            f"{tmp_dir}/{job_name}_{file_suffix}.json"
            if file_suffix
            else f"{tmp_dir}/{job_name}.json"
        )
        with open(file_path, "w") as f:
            f.write(saved_data)

        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            loaded_data = load_job_checkpoint(job_name, file_suffix)
            assert loaded_data == expected_checkpoint_data


@pytest.mark.xfail(raises=FileNotFoundError)
def test_load_job_checkpoint_raises_error_file_not_exists():
    job_name = "old_job"
    file_suffix = "correct_suffix"
    with tempfile.TemporaryDirectory() as tmp_dir:
        file_path = f"{tmp_dir}/{job_name}_{file_suffix}.json"
        with open(file_path, "w") as _:
            pass

        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            load_job_checkpoint(job_name, "wrong_suffix")


@pytest.mark.xfail(raises=ValueError)
def test_load_job_checkpoint_raises_error_corrupted_data():
    job_name = "old_job_corrupted_data"
    file_suffix = "foo"
    with tempfile.TemporaryDirectory() as tmp_dir:
        file_path = f"{tmp_dir}/{job_name}_{file_suffix}.json"
        with open(file_path, "w") as corrupted_file:
            corrupted_file.write(
                json.dumps(
                    {
                        "braketSchemaHeader": {
                            "name": "braket.jobs_data.persisted_job_data",
                            "version": "1",
                        },
                        "dataDictionary": {
                            "converged": "gASILg==\n",
                            "energy": "gASVCgBHv--corrupted---\n",
                        },
                        "dataFormat": "pickled_v4",
                    }
                )
            )

        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            load_job_checkpoint(job_name, file_suffix)


@dataclass
class CustomClassToPersist:
    float_val: float
    str_val: str
    bool_val: bool


def test_save_and_load_job_checkpoint():
    with tempfile.TemporaryDirectory() as tmp_dir:
        job_name = "job_name_1"
        data = {
            "np_array": np.array([1]),
            "custom_class": CustomClassToPersist(3.4, "str", True),
            "none_value": None,
            "nested_dict": {"a": {"b": False}},
        }
        with patch.dict(
            os.environ, {"AMZN_BRAKET_CHECKPOINT_DIR": tmp_dir, "AMZN_BRAKET_JOB_NAME": job_name}
        ):
            save_job_checkpoint(data, data_format=PersistedJobDataFormat.PICKLED_V4)
            retrieved = load_job_checkpoint(job_name)
            assert retrieved == data


@pytest.mark.parametrize(
    "data_format, result_data, expected_saved_data",
    [
        (
            PersistedJobDataFormat.PLAINTEXT,
            {"converged": True, "energy": -0.2},
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {"converged": True, "energy": -0.2},
                    "dataFormat": "plaintext",
                }
            ),
        ),
        (
            PersistedJobDataFormat.PICKLED_V4,
            {"converged": True, "energy": -0.2},
            json.dumps(
                {
                    "braketSchemaHeader": {
                        "name": "braket.jobs_data.persisted_job_data",
                        "version": "1",
                    },
                    "dataDictionary": {
                        "converged": "gASILg==\n",
                        "energy": "gASVCgAAAAAAAABHv8mZmZmZmZou\n",
                    },
                    "dataFormat": "pickled_v4",
                }
            ),
        ),
    ],
)
def test_save_job_result(data_format, result_data, expected_saved_data):
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(os.environ, {"AMZN_BRAKET_JOB_RESULTS_DIR": tmp_dir}):
            save_job_result(result_data, data_format)

        expected_file_location = f"{tmp_dir}/results.json"
        with open(expected_file_location, "r") as expected_file:
            assert expected_file.read() == expected_saved_data


@pytest.mark.parametrize("result_data", [{}, None])
def test_save_job_result_does_not_raise_error_empty_data(result_data):
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(os.environ, {"AMZN_BRAKET_JOB_RESULTS_DIR": tmp_dir}):
            save_job_result(result_data)


@pytest.mark.parametrize(
    "first_result_data,"
    "first_data_format,"
    "second_result_data,"
    "second_data_format,"
    "expected_result_data",
    (
        (
            "hello",
            PersistedJobDataFormat.PLAINTEXT,
            "goodbye",
            PersistedJobDataFormat.PLAINTEXT,
            {"result": "goodbye"},
        ),
        (
            "hello",
            PersistedJobDataFormat.PLAINTEXT,
            "goodbye",
            PersistedJobDataFormat.PICKLED_V4,
            {"result": "goodbye"},
        ),
        ("hello", PersistedJobDataFormat.PICKLED_V4, "goodbye", None, {"result": "goodbye"}),
        (
            # not json serializable
            PersistedJobDataFormat,
            PersistedJobDataFormat.PICKLED_V4,
            {"other_field": "value"},
            None,
            {"result": PersistedJobDataFormat, "other_field": "value"},
        ),
    ),
)
def test_update_result_data(
    first_result_data,
    first_data_format,
    second_result_data,
    second_data_format,
    expected_result_data,
):
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(os.environ, {"AMZN_BRAKET_JOB_RESULTS_DIR": tmp_dir}):
            save_job_result(first_result_data, first_data_format)
            save_job_result(second_result_data, second_data_format)

            assert load_job_result() == expected_result_data


def test_update_pickled_results_as_plaintext_error():
    with tempfile.TemporaryDirectory() as tmp_dir:
        with patch.dict(os.environ, {"AMZN_BRAKET_JOB_RESULTS_DIR": tmp_dir}):
            save_job_result(np.arange(5), PersistedJobDataFormat.PICKLED_V4)

            cannot_convert_pickled_to_plaintext = (
                "Cannot update results object serialized with "
                "pickled_v4 using data format plaintext."
            )
            with pytest.raises(TypeError, match=cannot_convert_pickled_to_plaintext):
                save_job_result("hello", PersistedJobDataFormat.PLAINTEXT)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from collections.abc import Sequence
from itertools import groupby
from typing import Any, Optional

from braket.circuits.basis_state import BasisState, BasisStateInput
from braket.circuits.quantum_operator import QuantumOperator
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    SerializationProperties,
)
from braket.registers.qubit_set import QubitSet


class Gate(QuantumOperator):
    """Class `Gate` represents a quantum gate that operates on N qubits. Gates are considered the
    building blocks of quantum circuits. This class is considered the gate definition containing
    the metadata that defines what a gate is and what it does.
    """

    def __init__(self, qubit_count: Optional[int], ascii_symbols: Sequence[str]):
        """Initializes a `Gate`.

        Args:
            qubit_count (Optional[int]): Number of qubits this gate interacts with.
            ascii_symbols (Sequence[str]): ASCII string symbols for the gate. These are used when
                printing a diagram of circuits. Length must be the same as `qubit_count`, and
                index ordering is expected to correlate with target ordering on the instruction.
                For instance, if CNOT instruction has the control qubit on the first index and
                target qubit on the second index. Then ASCII symbols would have ["C", "X"] to
                correlate a symbol with that index.

        Raises:
            ValueError: `qubit_count` is less than 1, `ascii_symbols` are `None`, or
                `ascii_symbols` length != `qubit_count`
        """
        # todo: implement ascii symbols for control modifier
        super().__init__(qubit_count=qubit_count, ascii_symbols=ascii_symbols)

    @property
    def _qasm_name(self) -> NotImplementedError:
        raise NotImplementedError()

    def adjoint(self) -> list[Gate]:
        """Returns a list of gates that implement the adjoint of this gate.

        This is a list because some gates do not have an inverse defined by a single existing gate.

        Returns:
            list[Gate]: The gates comprising the adjoint of this gate.
        """
        raise NotImplementedError(f"Gate {self.name} does not have adjoint implemented")

    def to_ir(
        self,
        target: QubitSet,
        ir_type: IRType = IRType.JAQCD,
        serialization_properties: Optional[SerializationProperties] = None,
        *,
        control: Optional[QubitSet] = None,
        control_state: Optional[BasisStateInput] = None,
        power: float = 1,
    ) -> Any:
        r"""Returns IR object of quantum operator and target

        Args:
            target (QubitSet): target qubit(s).
            ir_type(IRType) : The IRType to use for converting the gate object to its
                IR representation. Defaults to IRType.JAQCD.
            serialization_properties (Optional[SerializationProperties]): The serialization
                properties to use while serializing the object to the IR representation. The
                serialization properties supplied must correspond to the supplied `ir_type`.
                Defaults to None.
            control (Optional[QubitSet]): Control qubit(s). Only supported for OpenQASM.
                Default None.
            control_state (Optional[BasisStateInput]): Quantum state on which to control the
                operation. Must be a binary sequence of same length as number of qubits in
                `control`. Will be ignored if `control` is not present. May be represented as a
                string, list, or int. For example "0101", [0, 1, 0, 1], 5 all represent
                controlling on qubits 0 and 2 being in the \\|0⟩ state and qubits 1 and 3 being
                in the \\|1⟩ state. Default "1" * len(control).
            power (float): Integer or fractional power to raise the gate to. Negative
                powers will be split into an inverse, accompanied by the positive power.
                Default 1.

        Returns:
            Any: IR object of the quantum operator and target

        Raises:
            ValueError: If the supplied `ir_type` is not supported, or if the supplied serialization
            properties don't correspond to the `ir_type`.
            ValueError: If gate modifiers are supplied with `ir_type` Jaqcd.
        """
        if ir_type == IRType.JAQCD:
            if control or power != 1:
                raise ValueError("Gate modifiers are not supported with Jaqcd.")
            return self._to_jaqcd(target)
        elif ir_type == IRType.OPENQASM:
            if serialization_properties and not isinstance(
                serialization_properties, OpenQASMSerializationProperties
            ):
                raise ValueError(
                    "serialization_properties must be of type OpenQASMSerializationProperties "
                    "for IRType.OPENQASM."
                )
            return self._to_openqasm(
                target,
                serialization_properties or OpenQASMSerializationProperties(),
                control=control,
                control_state=control_state,
                power=power,
            )
        else:
            raise ValueError(f"Supplied ir_type {ir_type} is not supported.")

    def _to_jaqcd(self, target: QubitSet) -> Any:
        """Returns the JAQCD representation of the gate.

        Args:
            target (QubitSet): target qubit(s).

        Returns:
            Any: JAQCD object representing the gate.
        """
        raise NotImplementedError("to_jaqcd is not implemented.")

    def _to_openqasm(
        self,
        target: QubitSet,
        serialization_properties: OpenQASMSerializationProperties,
        *,
        control: Optional[QubitSet] = None,
        control_state: Optional[BasisStateInput] = None,
        power: float = 1,
    ) -> str:
        """Returns the OpenQASM string representation of the gate.

        Args:
            target (QubitSet): target qubit(s).
            serialization_properties (OpenQASMSerializationProperties): The serialization properties
                to use while serializing the object to the IR representation.
            control (Optional[QubitSet]): Control qubit(s). Default None.
            control_state (Optional[BasisStateInput]): Quantum state on which to control the
                operation. Must be a binary sequence of same length as number of qubits in
                `control`. Will be ignored if `control` is not present. May be represented as a
                string, list, or int. For example "0101", [0, 1, 0, 1], 5 all represent
                controlling on qubits 0 and 2 being in the \\|0⟩ state and qubits 1 and 3 being
                in the \\|1⟩ state. Default "1" * len(control).
            power (float): Integer or fractional power to raise the gate to. Negative
                powers will be split into an inverse, accompanied by the positive power.
                Default 1.

        Returns:
            str: Representing the openqasm representation of the gate.
        """
        target_qubits = [serialization_properties.format_target(int(qubit)) for qubit in target]
        if control:
            control_qubits = [
                serialization_properties.format_target(int(qubit)) for qubit in control
            ]
            control_state = (1,) * len(control) if control_state is None else control_state
            control_basis_state = BasisState(control_state, len(control))
            control_modifiers = []
            for state, group in groupby(control_basis_state.as_tuple):
                modifier_name = "neg" * (not state) + "ctrl"
                control_modifiers += [
                    (
                        f"{modifier_name}"
                        if (num_control := len(list(group))) == 1
                        else f"{modifier_name}({num_control})"
                    )
                ]
            control_modifiers.append("")
            qubits = control_qubits + target_qubits
            control_prefix = " @ ".join(control_modifiers)
        else:
            qubits = target_qubits
            control_prefix = ""
        inv_prefix = "inv @ " if power and power < 0 else ""
        power_prefix = f"pow({abs_power}) @ " if (abs_power := abs(power)) != 1 else ""
        param_string = (
            f"({', '.join(map(str, self.parameters))})" if hasattr(self, "parameters") else ""
        )

        return (
            f"{inv_prefix}{power_prefix}{control_prefix}"
            f"{self._qasm_name}{param_string}{','.join([f' {qubit}' for qubit in qubits])};"
        )

    @property
    def ascii_symbols(self) -> tuple[str, ...]:
        """tuple[str, ...]: Returns the ascii symbols for the quantum operator."""
        return self._ascii_symbols

    def __eq__(self, other: Gate):
        return isinstance(other, Gate) and self.name == other.name

    def __repr__(self):
        return f"{self.name}('qubit_count': {self._qubit_count})"

    def __hash__(self):
        return hash((self.name, self.qubit_count))

    @classmethod
    def register_gate(cls, gate: type[Gate]) -> None:
        """Register a gate implementation by adding it into the Gate class.

        Args:
            gate (type[Gate]): Gate class to register.
        """
        setattr(cls, gate.__name__, gate)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock

import numpy as np
import pytest

from braket.ahs.driving_field import DrivingField
from braket.ahs.field import Field
from braket.ahs.hamiltonian import Hamiltonian
from braket.timings.time_series import StitchBoundaryCondition, TimeSeries


@pytest.fixture
def default_driving_field():
    return DrivingField(Mock(spec=Field), Mock(spec=Field), Mock(spec=Field))


def test_create():
    mock0 = Mock(spec=Field)
    mock1 = Mock(spec=Field)
    mock2 = Mock(spec=Field)
    field = DrivingField(amplitude=mock0, phase=mock1, detuning=mock2)
    assert mock0 == field.amplitude
    assert mock1 == field.phase
    assert mock2 == field.detuning


def test_create_non_field():
    mock0 = Mock(spec=TimeSeries)
    mock1 = Mock(spec=TimeSeries)
    mock2 = Mock(spec=TimeSeries)
    field = DrivingField(amplitude=mock0, phase=mock1, detuning=mock2)
    assert mock0 == field.amplitude.time_series
    assert mock1 == field.phase.time_series
    assert mock2 == field.detuning.time_series


def test_add_hamiltonian(default_driving_field):
    expected = [default_driving_field, Mock(), Mock(), Mock()]
    result = expected[0] + Hamiltonian([expected[1], expected[2], expected[3]])
    assert result.terms == expected


def test_add_to_hamiltonian(default_driving_field):
    expected = [Mock(), Mock(), Mock(), default_driving_field]
    result = Hamiltonian([expected[0], expected[1], expected[2]]) + expected[3]
    assert result.terms == expected


def test_add_to_other():
    field0 = DrivingField(Mock(spec=Field), Mock(spec=Field), Mock(spec=Field))
    field1 = DrivingField(Mock(spec=Field), Mock(spec=Field), Mock(spec=Field))
    result = field0 + field1
    assert type(result) is Hamiltonian
    assert result.terms == [field0, field1]


def test_add_to_self(default_driving_field):
    result = default_driving_field + default_driving_field
    assert type(result) is Hamiltonian
    assert result.terms == [default_driving_field, default_driving_field]


def test_iadd_to_other(default_driving_field):
    expected = [Mock(), Mock(), Mock(), default_driving_field]
    other = Hamiltonian([expected[0], expected[1], expected[2]])
    other += expected[3]
    assert other.terms == expected


def test_discretize():
    amplitude_mock = Mock(spec=Field)
    amplitude_mock.discretize.return_value = Mock(spec=Field)
    phase_mock = Mock(spec=Field)
    phase_mock.discretize.return_value = Mock(spec=Field)
    detuning_mock = Mock(spec=Field)
    detuning_mock.discretize.return_value = Mock(spec=Field)
    mock_properties = Mock()
    field = DrivingField(amplitude=amplitude_mock, phase=phase_mock, detuning=detuning_mock)
    discretized_field = field.discretize(mock_properties)
    amplitude_mock.discretize.assert_called_with(
        time_resolution=mock_properties.rydberg.rydbergGlobal.timeResolution,
        value_resolution=mock_properties.rydberg.rydbergGlobal.rabiFrequencyResolution,
    )
    phase_mock.discretize.assert_called_with(
        time_resolution=mock_properties.rydberg.rydbergGlobal.timeResolution,
        value_resolution=mock_properties.rydberg.rydbergGlobal.phaseResolution,
    )
    detuning_mock.discretize.assert_called_with(
        time_resolution=mock_properties.rydberg.rydbergGlobal.timeResolution,
        value_resolution=mock_properties.rydberg.rydbergGlobal.detuningResolution,
    )
    assert field is not discretized_field
    assert discretized_field.amplitude == amplitude_mock.discretize.return_value
    assert discretized_field.phase == phase_mock.discretize.return_value
    assert discretized_field.detuning == detuning_mock.discretize.return_value


def test_from_lists():
    times = [0, 0.1, 0.2]
    amplitudes = [0.5, 0.8, 0.9]
    detunings = [0.3, 0.7, 0.6]
    phases = [0.2, 0.4, 0.6]

    dr_field = DrivingField.from_lists(times, amplitudes, detunings, phases)
    assert dr_field.amplitude.time_series.values() == amplitudes
    assert dr_field.detuning.time_series.values() == detunings
    assert dr_field.phase.time_series.values() == phases

    assert dr_field.amplitude.time_series.times() == times
    assert dr_field.detuning.time_series.times() == times
    assert dr_field.phase.time_series.times() == times


@pytest.mark.xfail(raises=ValueError)
def test_from_lists_not_eq_length():
    times = [0, 0.1]
    amplitudes = [0.5, 0.8, 0.9]
    detunings = [0.3, 0.7, 0.6]
    phases = [0.2, 0.4, 0.6]

    DrivingField.from_lists(times, amplitudes, detunings, phases)


def test_stitch():
    dr_field_1 = DrivingField.from_lists(
        times=[0, 0.1, 0.2],
        amplitudes=[1, 2, 3.5],
        detunings=[1.2, 3.4, 5.6],
        phases=[2.1, 4.2, 1.3],
    )
    dr_field_2 = DrivingField.from_lists(
        times=[0.4, 0.5, 0.6],
        amplitudes=[0.11, 0.22, 0.35],
        detunings=[1.12, 3.14, 5.16],
        phases=[2.11, 4.12, 1.13],
    )

    new_dr = dr_field_1.stitch(dr_field_2, boundary=StitchBoundaryCondition.RIGHT)
    new_times = new_dr.amplitude.time_series.times()

    amplitudes_1 = dr_field_1.amplitude.time_series.values()
    amplitudes_2 = dr_field_2.amplitude.time_series.values()
    new_amplitudes = new_dr.amplitude.time_series.values()

    detunings_1 = dr_field_1.detuning.time_series.values()
    detunings_2 = dr_field_2.detuning.time_series.values()
    new_detunings = new_dr.detuning.time_series.values()

    phases_1 = dr_field_1.phase.time_series.values()
    phases_2 = dr_field_2.phase.time_series.values()
    new_phases = new_dr.phase.time_series.values()

    expected_times = [0, 0.1, 0.2, 0.3, 0.4]
    np.testing.assert_almost_equal(new_times, expected_times)
    np.testing.assert_almost_equal(new_amplitudes, amplitudes_1[:-1] + amplitudes_2)
    np.testing.assert_almost_equal(new_detunings, detunings_1[:-1] + detunings_2)
    np.testing.assert_almost_equal(new_phases, phases_1[:-1] + phases_2)


@pytest.mark.xfail(raises=ValueError)
def test_iadd_to_itself(default_driving_field):
    default_driving_field += Hamiltonian(Mock())



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod

import braket.circuits.circuit as cir


class CircuitDiagram(ABC):
    """A class that builds circuit diagrams."""

    @staticmethod
    @abstractmethod
    def build_diagram(circuit: cir.Circuit) -> str:
        """Build a diagram for the specified `circuit`.

        Args:
            circuit (cir.Circuit): The circuit to build a diagram for.

        Returns:
            str: String representation for the circuit diagram.
            An empty string is returned for an empty circuit.
        """



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import sys
from unittest.mock import Mock

import braket.ipython_utils as ipython_utils


def test_running_in_jupyter():
    assert not ipython_utils.running_in_jupyter()


def test_ipython_imported_but_ipython_none():
    _mock_ipython(None)
    assert not ipython_utils.running_in_jupyter()


def test_ipython_imported_but_not_in_jupyter():
    _mock_ipython(MockIPython(None))
    assert not ipython_utils.running_in_jupyter()


def test_ipython_imported_and_in_jupyter():
    _mock_ipython(MockIPython("non empty kernel"))
    assert ipython_utils.running_in_jupyter()


def get_ipython():
    pass


def _mock_ipython(get_ipython_result):
    module = sys.modules["test_ipython_utils"]
    sys.modules["IPython"] = module

    get_ipython = Mock(return_value=get_ipython_result)
    sys.modules["IPython"].__dict__["get_ipython"] = get_ipython


class MockIPython:
    def __init__(self, kernel):
        self.kernel = kernel



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import Gate, QuantumOperator, QubitSet
from braket.circuits.serialization import IRType


@pytest.fixture
def gate():
    return Gate(qubit_count=1, ascii_symbols=["foo"])


def test_is_operator(gate):
    assert isinstance(gate, QuantumOperator)


def test_adjoint_not_implemented_by_default(gate):
    with pytest.raises(NotImplementedError):
        gate.adjoint()


def test_to_matrix_not_implemented_by_default(gate):
    with pytest.raises(NotImplementedError):
        gate.to_matrix(None)


def test_matrix_equivalence():
    gate1 = Gate.H()
    gate2 = Gate.H()
    gate3 = Gate.CNot()
    assert gate1.matrix_equivalence(gate2)
    assert not gate2.matrix_equivalence(gate3)


def test_matrix_equivalence_non_gate():
    gate1 = Gate.H()
    assert not gate1.matrix_equivalence(1)


def test_str(gate):
    expected = f"{gate.name}('qubit_count': {gate.qubit_count})"
    assert str(gate) == expected


def test_str_angle():
    gate = Gate.Rx(0.5)
    expected = f"{gate.name}('angle': {gate.angle}, 'qubit_count': {gate.qubit_count})"
    assert str(gate) == expected


def test_equality():
    gate_1 = Gate(qubit_count=1, ascii_symbols=["foo"])
    gate_2 = Gate(qubit_count=1, ascii_symbols=["bar"])
    other_gate = Gate.Rx(angle=0.34)
    non_gate = "non gate"

    assert gate_1 == gate_2
    assert gate_1 is not gate_2
    assert gate_1 != other_gate
    assert gate_1 != non_gate


def test_register_gate():
    class _FooGate(Gate):
        def __init__(self):
            super().__init__(qubit_count=1, ascii_symbols=["foo"])

    Gate.register_gate(_FooGate)
    assert Gate._FooGate().name == _FooGate().name


@pytest.mark.parametrize(
    "ir_type, serialization_properties, expected_exception, expected_message, control",
    [
        (IRType.JAQCD, None, NotImplementedError, "to_jaqcd is not implemented.", None),
        (
            IRType.JAQCD,
            None,
            ValueError,
            "Gate modifiers are not supported with Jaqcd.",
            QubitSet(0),
        ),
        (
            "invalid-ir-type",
            None,
            ValueError,
            "Supplied ir_type invalid-ir-type is not supported.",
            None,
        ),
        (
            IRType.OPENQASM,
            "invalid-property-type",
            ValueError,
            "serialization_properties must be of type OpenQASMSerializationProperties for "
            "IRType.OPENQASM.",
            None,
        ),
    ],
)
def test_gate_to_ir(
    ir_type, serialization_properties, expected_exception, expected_message, gate, control
):
    with pytest.raises(expected_exception) as exc:
        gate.to_ir(
            QubitSet(0), ir_type, serialization_properties=serialization_properties, control=control
        )
    assert exc.value.args[0] == expected_message



from __future__ import annotations

from functools import singledispatch
from typing import Optional, Union

import numpy as np


class BasisState:
    def __init__(self, state: BasisStateInput, size: Optional[int] = None):
        self.state = _as_tuple(state, size)

    @property
    def size(self) -> int:
        return len(self.state)

    @property
    def as_tuple(self) -> tuple:
        return self.state

    @property
    def as_int(self) -> int:
        return 2 ** np.arange(self.size)[::-1] @ self.state

    @property
    def as_string(self) -> str:
        return "".join(map(str, self.state))

    def __len__(self) -> int:
        return len(self.state)

    def __iter__(self):
        return iter(self.state)

    def __eq__(self, other: BasisState):
        return self.state == other.state

    def __bool__(self):
        return any(self.state)

    def __str__(self):
        return self.as_string

    def __repr__(self):
        return f'BasisState("{self.as_string}")'

    def __getitem__(self, item: int):
        return BasisState(self.state[item])


BasisStateInput = Union[int, list[int], str, BasisState]


@singledispatch
def _as_tuple(state: BasisStateInput, size: int) -> tuple:
    size = size if size is not None else len(state)
    if state and len(state) > size:
        raise ValueError(
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits."
        )
    return (0,) * (size - len(state)) + tuple(state)


@_as_tuple.register
def _(state: int, size: int):
    if size is not None and state >= 2**size:
        raise ValueError(
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits."
        )
    return tuple(int(x) for x in np.binary_repr(state, size))


@_as_tuple.register
def _(state: str, size: int):
    size = size if size is not None else len(state)
    if len(state) > size:
        raise ValueError(
            "State value represents a binary sequence of length greater "
            "than the specified number of qubits."
        )
    # left-pad to match state size
    return (0,) * (size - len(state)) + tuple(int(x) for x in state)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock

from braket.ahs.hamiltonian import Hamiltonian


def test_create():
    hamiltonian = Hamiltonian()
    assert hamiltonian.terms == []


def test_add():
    mocks = [Mock(), Mock(), Mock(), Mock()]
    result = Hamiltonian([mocks[0], mocks[1], mocks[2]]) + Hamiltonian([mocks[3]])
    assert result.terms == mocks


def test_iadd():
    mocks = [Mock(), Mock(), Mock(), Mock()]
    hamiltonian = Hamiltonian([mocks[0]])
    hamiltonian += Hamiltonian([mocks[1], mocks[2], mocks[3]])
    assert hamiltonian.terms == mocks


def test_discretize():
    mocks = [Mock(), Mock(), Mock()]
    mock_properties = Mock()
    hamiltonian = Hamiltonian(mocks)
    discretized_hamiltonian = hamiltonian.discretize(mock_properties)
    for index in range(len(mocks)):
        mocks[index].discretize.assert_called_with(mock_properties)
        assert discretized_hamiltonian.terms[index] == mocks[index].discretize.return_value
    assert hamiltonian is not discretized_hamiltonian



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from typing import Optional

from braket.ahs.discretization_types import DiscretizationProperties


class Hamiltonian:
    def __init__(self, terms: Optional[list[Hamiltonian]] = None):
        r"""A Hamiltonian representing a system to be simulated.

        A Hamiltonian :math:`H` may be expressed as a sum of multiple terms

        .. math::
            H = \sum_i H_i
        """
        self._terms = terms or []

    @property
    def terms(self) -> list[Hamiltonian]:
        """list[Hamiltonian]: The list of terms in this Hamiltonian."""
        return self._terms

    def discretize(self, properties: DiscretizationProperties) -> Hamiltonian:
        """Creates a discretized version of the Hamiltonian.

        Args:
            properties (DiscretizationProperties): Capabilities of a device that represent the
                resolution with which the device can implement the parameters.

        Returns:
            Hamiltonian: A new discretized Hamiltonian.
        """
        terms = [term.discretize(properties) for term in self.terms]
        return Hamiltonian(terms=terms)

    def __iadd__(self, other: Hamiltonian) -> Hamiltonian:
        if type(self) is not Hamiltonian:
            raise ValueError(f"Unable to modify Hamiltonian of type {type(self)}")
        self._terms.extend(other.terms)
        return self

    def __add__(self, other: Hamiltonian) -> Hamiltonian:
        terms = []
        terms.extend(self.terms)
        terms.extend(other.terms)
        return Hamiltonian(terms)
